# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform model endpoint operations."""
from __future__ import annotations

from typing import Any, Final, cast

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from sqlalchemy.orm import aliased
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.artifacts.service import ArtifactIdService
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

MODEL_RESOURCE_TYPE: Final[str] = "ml_model"
MODEL_VERSION_RESOURCE_TYPE: Final[str] = "ml_model_version"
MODEL_SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.MlModel.name.like(x, escape="/"),
    "description": lambda x: models.MlModel.description.like(x, escape="/"),
}
MODEL_VERSION_SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "description": lambda x: models.MlModelVersion.description.like(x, escape="/"),
}
MODEL_SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.MlModel.name,
    "createdOn": models.MlModel.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.MlModel.description,
}


class ModelService(object):
    """The service methods for registering and managing models by their unique id."""

    @inject
    def __init__(
        self,
        model_name_service: ModelNameService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the model service.

        All arguments are provided via dependency injection.

        Args:
            model_name_service: A ModelNameService object.
            group_id_service: A GroupIdService object.
        """
        self._model_name_service = model_name_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        description: str,
        group_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.ModelWithVersionDict:
        """Create a new model.

        Args:
            name: The name of the model. The combination of name and group_id must be
                unique.
            description: The description of the model.
            group_id: The group that will own the model.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created model object.

        Raises:
            EntityExistsError: If a model with the given name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._model_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                MODEL_RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)

        resource = models.Resource(resource_type=MODEL_RESOURCE_TYPE, owner=group)

        ml_model = models.MlModel(
            name=name,
            description=description,
            resource=resource,
            creator=current_user,
        )

        db.session.add(ml_model)

        if commit:
            db.session.commit()
            log.debug(
                "Model registration successful",
                model_id=ml_model.resource_id,
                name=ml_model.name,
            )

        return utils.ModelWithVersionDict(
            ml_model=ml_model, version=None, has_draft=False
        )

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> Any:
        """Fetch a list of models, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of models to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of models and the total number of models matching
            the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when counting
                the number of models.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of models")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, MODEL_SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.MlModel.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModel.resource_snapshot_id,
            )
        )
        total_num_ml_models = db.session.scalars(stmt).first()

        if total_num_ml_models is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_ml_models == 0:
            return [], total_num_ml_models

        latest_ml_models_stmt = (
            select(models.MlModel)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModel.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in MODEL_SORTABLE_FIELDS:
            sort_column = MODEL_SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            latest_ml_models_stmt = latest_ml_models_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in MODEL_SORTABLE_FIELDS:
            raise SortParameterValidationError(MODEL_RESOURCE_TYPE, sort_by_string)

        ml_models = db.session.scalars(latest_ml_models_stmt).all()

        # extract list of model ids
        model_ids = [ml_model.resource_id for ml_model in ml_models]

        # Build CTE that retrieves all snapshot ids for the list of ml model versions
        # associated with retrieved ml models
        parent_model = aliased(models.MlModel)
        model_version_snapshot_ids_cte = (
            select(models.MlModelVersion.resource_snapshot_id)
            .join(
                models.resource_dependencies_table,
                models.MlModelVersion.resource_id
                == models.resource_dependencies_table.c.child_resource_id,
            )
            .join(
                parent_model,
                parent_model.resource_id
                == models.resource_dependencies_table.c.parent_resource_id,
            )
            .where(parent_model.resource_id.in_(model_ids))
            .cte()
        )

        # get the latest model version snapshots associated with the retrieved models
        latest_model_versions_stmt = (
            select(models.MlModelVersion)
            .join(models.Resource)
            .where(
                models.MlModelVersion.resource_snapshot_id.in_(
                    select(model_version_snapshot_ids_cte)
                ),
                models.Resource.latest_snapshot_id
                == models.MlModelVersion.resource_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
            .order_by(models.MlModelVersion.created_on)
        )
        model_versions = db.session.scalars(latest_model_versions_stmt).unique().all()

        # build a dictionary structure to re-associate models with model versions
        models_dict: dict[int, utils.ModelWithVersionDict] = {
            model.resource_id: utils.ModelWithVersionDict(
                ml_model=model, version=None, has_draft=False
            )
            for model in ml_models
        }
        for model_version in model_versions:
            models_dict[model_version.model_id]["version"] = model_version

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(
                tuple(model["ml_model"].resource_id for model in models_dict.values())
            ),
            models.DraftResource.user_id == current_user.user_id,
        )
        for resource_id in db.session.scalars(drafts_stmt):
            models_dict[resource_id]["has_draft"] = True

        return list(models_dict.values()), total_num_ml_models


class ModelIdService(object):
    """The service methods for registering and managing models by their unique id."""

    @inject
    def __init__(
        self,
        model_name_service: ModelNameService,
    ) -> None:
        """Initialize the model service.

        All arguments are provided via dependency injection.

        Args:
            model_name_service: A ModelNameService object.
        """
        self._model_name_service = model_name_service

    def get(
        self,
        model_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.ModelWithVersionDict | None:
        """Fetch a model by its unique id.

        Args:
            model_id: The unique id of the model.
            error_if_not_found: If True, raise an error if the model is not found.
                Defaults to False.

        Returns:
            The model object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the model is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get model by id", model_id=model_id)

        stmt = (
            select(models.MlModel)
            .join(models.Resource)
            .where(
                models.MlModel.resource_id == model_id,
                models.MlModel.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        ml_model = db.session.scalars(stmt).first()

        if ml_model is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(MODEL_RESOURCE_TYPE, model_id=model_id)

            return None

        latest_ml_model_version_stmt = (
            select(models.MlModelVersion)
            .join(models.Resource)
            .where(
                models.MlModelVersion.model_id == model_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModelVersion.resource_snapshot_id,
            )
            .order_by(models.MlModelVersion.created_on.desc())
            .limit(1)
        )
        latest_version = db.session.scalar(latest_ml_model_version_stmt)

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == ml_model.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.ModelWithVersionDict(
            ml_model=ml_model, version=latest_version, has_draft=has_draft
        )

    def modify(
        self,
        model_id: int,
        name: str,
        description: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.ModelWithVersionDict | None:
        """Modify a model.

        Args:
            model_id: The unique id of the model.
            name: The new name of the model.
            description: The new description of the model.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated model object.

        Raises:
            EntityDoesNotExistError: If the model is not found and `error_if_not_found`
                is True.
            EntityExistsError: If the model name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        ml_model_dict = self.get(
            model_id, error_if_not_found=error_if_not_found, log=log
        )

        if ml_model_dict is None:
            return None

        ml_model = ml_model_dict["ml_model"]
        version = ml_model_dict["version"]
        has_draft = ml_model_dict["has_draft"]

        group_id = ml_model.resource.group_id
        if name != ml_model.name:
            duplicate = self._model_name_service.get(name, group_id=group_id, log=log)
            if duplicate is not None:
                raise EntityExistsError(
                    MODEL_RESOURCE_TYPE,
                    duplicate.resource_id,
                    name=name,
                    group_id=group_id,
                )

        new_ml_model = models.MlModel(
            name=name,
            description=description,
            resource=ml_model.resource,
            creator=current_user,
        )
        db.session.add(new_ml_model)

        if commit:
            db.session.commit()
            log.debug(
                "Model modification successful",
                model_id=model_id,
                name=name,
                description=description,
            )

        return utils.ModelWithVersionDict(
            ml_model=new_ml_model,
            version=version,
            has_draft=has_draft,
        )

    def delete(self, model_id: int, **kwargs) -> dict[str, Any]:
        """Delete a model.

        Args:
            model_id: The unique id of the model.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=model_id, resource_type=MODEL_RESOURCE_TYPE, is_deleted=False
        )
        model_resource = db.session.scalars(stmt).first()

        if model_resource is None:
            raise EntityDoesNotExistError(MODEL_RESOURCE_TYPE, model_id=model_id)

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type="delete",
            resource=model_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug("Model deleted", model_id=model_id)

        return {"status": "Success", "model_id": model_id}


class ModelIdVersionsService(object):
    @inject
    def __init__(
        self,
        artifact_id_service: ArtifactIdService,
        model_id_service: ModelIdService,
    ) -> None:
        """Initialize the model service.

        All arguments are provided via dependency injection.

        Args:
            artifact_id_service: A ArtifactIdService object.
            model_id_service: A ModelIdService object.
        """
        self._artifact_id_service = artifact_id_service
        self._model_id_service = model_id_service

    def create(
        self,
        model_id: int,
        description: str,
        artifact_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.ModelWithVersionDict:
        """Create a new model version.

        Args:
            model_id: The unique id of the model.
            description: The description of the model version.
            artifact_id: The artifact for the model version.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created model object.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        ml_model_dict = cast(
            utils.ModelWithVersionDict,
            self._model_id_service.get(model_id, error_if_not_found=True, log=log),
        )
        ml_model = ml_model_dict["ml_model"]
        has_draft = ml_model_dict["has_draft"]
        version = ml_model_dict["version"]
        next_version_number = version.version_number + 1 if version is not None else 1
        group = ml_model.resource.owner
        artifact_dict = cast(
            utils.ArtifactDict,
            self._artifact_id_service.get(
                artifact_id, error_if_not_found=True, log=log
            ),
        )
        artifact = artifact_dict["artifact"]

        resource = models.Resource(
            resource_type=MODEL_VERSION_RESOURCE_TYPE, owner=group
        )
        new_version = models.MlModelVersion(
            description=description,
            artifact=artifact,
            version_number=next_version_number,
            resource=resource,
            creator=current_user,
        )

        ml_model.resource.children.append(new_version.resource)
        db.session.add(new_version)

        if commit:
            db.session.commit()
            log.debug(
                "Model registration successful",
                model_id=ml_model.resource_id,
                name=ml_model.name,
            )

        return utils.ModelWithVersionDict(
            ml_model=ml_model,
            version=new_version,
            has_draft=has_draft,
        )

    def get(
        self,
        model_id: int,
        search_string: str,
        page_index: int,
        page_length: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> tuple[list[utils.ModelWithVersionDict], int] | None:
        """Fetch a list of versions of a model.

        Args:
            model_id: The unique id of the resource.
            search_string: A search string used to filter results.
            page_index: The index of the first snapshot to be returned.
            page_length: The maximum number of versions to be returned.
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.

        Returns:
            The list of resource versions of the resource object if found, otherwise
                None.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get model versions by id", model_id=model_id)

        ml_model_dict = cast(
            utils.ModelWithVersionDict,
            self._model_id_service.get(model_id, error_if_not_found=True, log=log),
        )
        ml_model = ml_model_dict["ml_model"]
        model_id = ml_model.resource_id

        filters = construct_sql_query_filters(
            search_string, MODEL_VERSION_SEARCHABLE_FIELDS
        )

        latest_model_versions_count_stmt = (
            select(func.count(models.MlModelVersion.resource_id))
            .join(models.Resource)
            .where(
                filters,
                models.MlModelVersion.model_id == model_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModelVersion.resource_snapshot_id,
            )
        )
        total_num_model_versions = db.session.scalar(latest_model_versions_count_stmt)

        if total_num_model_versions is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(total_num_model_versions),
            )
            raise BackendDatabaseError

        if total_num_model_versions == 0:
            return [], total_num_model_versions

        latest_model_versions_stmt = (
            select(models.MlModelVersion)
            .join(models.Resource)
            .where(
                filters,
                models.MlModelVersion.model_id == model_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModelVersion.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        drafts_stmt = (
            select(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
            )
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == model_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return [
            utils.ModelWithVersionDict(
                ml_model=ml_model, version=version, has_draft=has_draft
            )
            for version in db.session.scalars(latest_model_versions_stmt).unique().all()
        ], total_num_model_versions


class ModelIdVersionsNumberService(object):
    @inject
    def __init__(
        self,
        model_id_service: ModelIdService,
    ) -> None:
        """Initialize the model service.

        All arguments are provided via dependency injection.

        Args:
            model_id_service: A ModelIdService object.
        """
        self._model_id_service = model_id_service

    def get(
        self,
        model_id: int,
        version_number: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.ModelWithVersionDict | None:
        """Fetch a specific version of a Model resource.

        Args:
            model_id: The unique id of the Model resource.
            version_number: The version number of the Model
            error_if_not_found: If True, raise an error if the resource is not found.
                Defaults to False.

        Returns:
            The requested version the resource object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the resource is not found and
                `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get resource snapshot by id", model_id=model_id)

        ml_model_dict = self._model_id_service.get(
            model_id, error_if_not_found=error_if_not_found, log=log
        )

        if ml_model_dict is None:
            return None

        ml_model = ml_model_dict["ml_model"]

        ml_model_version_stmt = (
            select(models.MlModelVersion)
            .join(models.Resource)
            .where(
                models.MlModelVersion.model_id == model_id,
                models.MlModelVersion.version_number == version_number,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModelVersion.resource_snapshot_id,
            )
        )
        latest_version = db.session.scalar(ml_model_version_stmt)

        if latest_version is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    MODEL_VERSION_RESOURCE_TYPE,
                    model_id=model_id,
                    version_number=version_number,
                )

            return None

        return utils.ModelWithVersionDict(
            ml_model=ml_model, version=latest_version, has_draft=False
        )

    def modify(
        self,
        model_id: int,
        version_number: int,
        description: str,
        commit: bool = True,
        **kwargs,
    ) -> utils.ModelWithVersionDict | None:
        """Modify a model version.

        Args:
            model_id: The unique id of the model.
            version_number: The version number of the model.
            description: The new description of the model version.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated model object.

        Raises:
            EntityDoesNotExistError: If the model is not found and `error_if_not_found`
                is True.
            EntityExistsError: If the model name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        ml_model_dict = cast(
            utils.ModelWithVersionDict,
            self.get(model_id, version_number, error_if_not_found=True, log=log),
        )

        ml_model = ml_model_dict["ml_model"]
        version = cast(models.MlModelVersion, ml_model_dict["version"])

        new_version = models.MlModelVersion(
            description=description,
            artifact=version.artifact,
            version_number=version.version_number,
            resource=version.resource,
            creator=current_user,
        )
        db.session.add(new_version)

        if commit:
            db.session.commit()
            log.debug(
                "Model Version modification successful",
                model_id=model_id,
                version_number=version_number,
                description=description,
            )

        return utils.ModelWithVersionDict(
            ml_model=ml_model,
            version=new_version,
            has_draft=False,
        )


class ModelNameService(object):
    """The service methods for managing models by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.MlModel | None:
        """Fetch a model by its name.

        Args:
            name: The name of the model.
            group_id: The the group id of the model.
            error_if_not_found: If True, raise an error if the model is not found.
                Defaults to False.

        Returns:
            The model object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the model is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get model by name", model_name=name, group_id=group_id)

        stmt = (
            select(models.MlModel)
            .join(models.Resource)
            .where(
                models.MlModel.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.MlModel.resource_snapshot_id,
            )
        )
        ml_model = db.session.scalars(stmt).first()

        if ml_model is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    MODEL_RESOURCE_TYPE, name=name, group_id=group_id
                )

            return None

        return ml_model



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import json
import pathlib
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Union

from dioptra.sdk.exceptions.base import BaseTaskEngineError
from dioptra.task_engine import type_registry, type_validation, types, util
from dioptra.task_engine.error_message import json_path_to_string
from dioptra.task_engine.issues import IssueSeverity, IssueType, ValidationIssue

_SCHEMA_FILENAME = "experiment_schema.json"


def _instance_path_to_description(  # noqa: C901
    instance_path: Sequence[Union[int, str]]
) -> str:
    """
    Create a nice description of the location in an experiment description
    pointed to by instance_path.  This implementation is crafted specifically
    to the structure of a declarative experiment description.

    Args:
        instance_path: A path, as a sequence of strings/ints.

    Returns:
        A string description
    """

    path_len = len(instance_path)

    message_parts = []
    if path_len == 0:
        message_parts.append("root level of experiment description")

    else:
        if instance_path[0] == "types":
            if path_len == 1:
                message_parts.append("types section")
            else:
                message_parts.append('definition of type "{}"'.format(instance_path[1]))

                if path_len > 2:
                    # Can't slice a Sequence! :-P
                    def_loc = list(instance_path)[2:]
                    message_parts.append("at location " + json_path_to_string(def_loc))

        elif instance_path[0] == "parameters":
            if path_len == 1:
                message_parts.append("global parameters section")
            else:
                message_parts.append("parameter")
                message_parts.append('"{}"'.format(instance_path[1]))

        elif instance_path[0] == "tasks":
            if path_len == 1:
                message_parts.append("tasks section")
            else:
                message_parts.append('task plugin "' + str(instance_path[1]) + '"')
                if path_len > 2:
                    if instance_path[2] == "outputs":
                        message_parts.append("outputs")
                    elif instance_path[2] == "inputs":
                        message_parts.append("inputs")
                    elif instance_path[2] == "plugin":
                        message_parts.append("plugin ID")

        elif instance_path[0] == "graph":
            if path_len == 1:
                message_parts.append("graph section")
            else:
                message_parts.append('step "' + str(instance_path[1]) + '"')
                if len(instance_path) > 2 and instance_path[2] == "dependencies":
                    message_parts.append("dependencies")

    if message_parts:
        description = " ".join(message_parts)
    else:
        # fallbacks if we don't know another way of describing the location
        instance_path_str = json_path_to_string(instance_path)
        description = "experiment description location " + instance_path_str

    return description


def _get_json_schema() -> Union[dict, bool]:  # hypothetical types of schemas
    """
    Read and parse the declarative experiment description JSON-Schema file.

    Returns:
        The schema, as parsed JSON
    """
    # Currently assumes the schema json file and this source file are in the
    # same directory.
    schema_path = pathlib.Path(__file__).with_name(_SCHEMA_FILENAME)

    schema: Union[dict, bool]
    with schema_path.open("r", encoding="utf-8") as fp:
        schema = json.load(fp)

    return schema


def _schema_validate(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Validate the given declarative experiment description against a JSON-Schema
    schema.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    schema = _get_json_schema()

    error_messages = util.schema_validate(
        experiment_desc, schema, _instance_path_to_description
    )

    issues = [
        ValidationIssue(IssueType.SCHEMA, IssueSeverity.ERROR, message)
        for message in error_messages
    ]

    return issues


def _find_non_string_keys(mapping: Mapping, key_meaning: str) -> list[ValidationIssue]:
    """
    Look for non-string keys in the given mapping.  Create validation issues
    for any non-string keys found.

    Args:
        mapping: A mapping to check
        key_meaning: Used to describe the keys being checked, in a validation
            issue message.  This way, we can use the same code to check keys
            in several places, but get issue messages which are not generic:
            they describe the role those keys play in the experiment
            description.

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    issues = []
    for key in mapping:
        message = None
        if isinstance(key, str):
            if not key or key.isspace():
                message = "{} names must be non-empty/whitespace: {!r}".format(
                    key_meaning, key
                )
        else:
            message = "{} must be strings: {!r}".format(key_meaning, key)

        if message:
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

    return issues


def _check_string_keys(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Ensure certain mappings within the experiment description use only string
    keys.  This check targets certain mappings: type names, global parameter
    names, task short names, graph step names.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    types = experiment_desc.get("types", {})
    parameters = experiment_desc.get("parameters", {})
    tasks = experiment_desc["tasks"]
    graph = experiment_desc["graph"]
    issues = []

    issues += _find_non_string_keys(types, "Type names")
    issues += _find_non_string_keys(parameters, "Global parameter names")
    issues += _find_non_string_keys(tasks, "Task short names")
    issues += _find_non_string_keys(graph, "Graph step names")

    return issues


def _check_name_collisions(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Check whether any graph step names collide with any parameter names.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    param_names = set(experiment_desc.get("parameters", []))
    step_names = set(experiment_desc["graph"])

    collisions = param_names & step_names

    issues = []
    if collisions:
        message = "Some parameters and steps have the same name: " + ", ".join(
            collisions
        )
        issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
        issues.append(issue)

    return issues


def _check_global_parameter_types(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check whether all global parameter types are valid.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    global_parameters = experiment_desc.get("parameters", {})
    type_defs = experiment_desc.get("types", {})
    type_names = type_registry.BUILTIN_TYPES.keys() | type_defs.keys()

    issues = []

    # Can't really use None to mean no default, since that's a valid default!
    no_default = object()

    for param_name, param_def in global_parameters.items():
        if isinstance(param_def, Mapping):
            param_type = param_def.get("type")
            param_default = param_def.get("default", no_default)

            if param_type is not None and param_type not in type_names:
                message = 'In global parameter "{}": undefined type: {}'.format(
                    param_name, param_type
                )

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

            if param_type is None and param_default is no_default:
                # Neither a type nor a default value was given.  This is an
                # invalid situation since there is no way to obtain a parameter
                # type.  I don't think you could create this situation and still
                # be schema-valid, but I want to keep the check anyway.

                message = (
                    'In global parameter "{}": unable to obtain a type:'
                    " neither a type nor default value were given"
                ).format(param_name)

                issue = ValidationIssue(IssueType.TYPE, IssueSeverity.ERROR, message)

                issues.append(issue)

    return issues


def _check_type_definition_type_references(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check for references to undefined types in all type definitions.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    type_defs = experiment_desc.get("types", {})
    type_names = type_registry.BUILTIN_TYPES.keys() | type_defs.keys()

    issues = []

    for type_name, type_def in type_defs.items():
        for type_ref in type_registry.get_dependency_types(type_def):
            if type_ref not in type_names:
                message = (
                    'In definition of type "{}": reference to undefined type: {}'
                ).format(type_name, type_ref)

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

    return issues


def _check_type_reference_cycle(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check for a reference cycle among type definitions.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    type_defs = experiment_desc.get("types", {})
    issues = []
    message = None

    try:
        type_registry.get_sorted_types(type_defs)

    except BaseTaskEngineError as e:
        # If all references resolve and the description is schema-valid, the
        # only exception that could be thrown is a TypeReferenceCycleError.
        # But I will catch all task engine errors, just in case.
        message = str(e)

    if message:
        issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
        issues.append(issue)

    return issues


def _check_union_member_duplicates(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check for union type definitions for which there is duplication in the
    membership.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    issues = []
    type_defs = experiment_desc.get("types", {})
    type_reg = type_registry.build_type_registry(type_defs)

    for type_name, type_ in type_reg.items():
        if isinstance(type_, types.UnionType):
            # Deduping automatically happens when union types are built.
            # So we can't just check a union type object for dupes; there will
            # never be any.  We must re-create every member type from the
            # definition in order to tell whether there were any duplicates.

            union_type_def = type_defs[type_name]
            member_type_defs = union_type_def["union"]

            dupe_types = set()
            seen_types = set()
            for member_type_def in member_type_defs:
                member_type = type_registry.build_or_get_type(member_type_def, type_reg)

                if member_type in seen_types:
                    dupe_types.add(member_type)
                else:
                    seen_types.add(member_type)

            if dupe_types:
                message = (
                    'In definition of type "{}": duplicate member types'
                    " encountered: {}"
                ).format(type_name, ", ".join(str(t) for t in dupe_types))

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.WARNING, message
                )

                issues.append(issue)

    return issues


def _check_task_plugin_references(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check whether all task plugin short names refer to known task plugins.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    task_defs = experiment_desc["tasks"]
    graph = experiment_desc["graph"]

    issues = []
    for step_name, step_def in graph.items():
        task_plugin_short_name = util.step_get_plugin_short_name(step_def)

        if task_plugin_short_name not in task_defs:
            message = 'In step "{}": unrecognized task plugin: {}'.format(
                step_name, task_plugin_short_name
            )
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

    return issues


def _check_task_plugin_pyplugs_coords(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check task plugin IDs for validity.  They must at minimum include a module
    name and a function name.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    task_defs = experiment_desc["tasks"]

    issues = []
    for task_short_name, task_def in task_defs.items():
        plugin = task_def["plugin"]
        if "." not in plugin:
            message = 'In task "{}": plugin ID requires at least one ".": {}'.format(
                task_short_name, plugin
            )
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

        plugin_parts = plugin.split(".")
        if any(len(plugin_part) == 0 for plugin_part in plugin_parts):
            message = (
                'In task "{}": all plugin module components must be of'
                " non-zero length: {}"
            ).format(task_short_name, plugin)
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

    return issues


def _check_task_plugin_io_names(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check task definitions for duplicate input and output names.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    task_defs = experiment_desc["tasks"]
    issues = []

    for short_task_name, task_def in task_defs.items():
        inputs = task_def.get("inputs", [])
        names = set()
        repeated_names = set()

        for input_ in inputs:
            name, _ = util.input_def_get_name_type(input_)
            if name in names:
                repeated_names.add(name)
            else:
                names.add(name)

        if repeated_names:
            message = 'In task "{}": some input names are repeated: {}'.format(
                short_task_name, ", ".join(repeated_names)
            )

            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)

            issues.append(issue)

        outputs = task_def.get("outputs", [])
        repeated_names.clear()
        names.clear()

        if not isinstance(outputs, list):
            outputs = [outputs]

        for output in outputs:
            name = next(iter(output))
            if name in names:
                repeated_names.add(name)
            else:
                names.add(name)

        if repeated_names:
            message = 'In task "{}": some output names are repeated: {}'.format(
                short_task_name, ", ".join(repeated_names)
            )

            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)

            issues.append(issue)

    return issues


def _check_task_plugin_io_types(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check task definition input and output type names for validity: whether
    they name known types.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    type_defs = experiment_desc.get("types", {})
    task_defs = experiment_desc["tasks"]
    issues = []

    type_names = type_registry.BUILTIN_TYPES.keys() | type_defs.keys()

    for short_task_name, task_def in task_defs.items():
        inputs = task_def.get("inputs", [])

        for input_ in inputs:
            name, type_ = util.input_def_get_name_type(input_)
            if type_ not in type_names:
                message = 'In task "{}": input "{}" has undefined type: {}'.format(
                    short_task_name, name, type_
                )

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

        outputs = task_def.get("outputs", [])

        if not isinstance(outputs, list):
            outputs = [outputs]

        for output in outputs:
            name, type_ = next(iter(output.items()))
            if type_ not in type_names:
                message = 'In task "{}": output "{}" has undefined type: {}'.format(
                    short_task_name, name, type_
                )

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

    return issues


def _check_graph_references(  # noqa: C901
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Scan for references within task invocations, check whether they are legal,
    and whether they refer to recognized parameters, steps, and/or step outputs.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    graph = experiment_desc["graph"]
    task_defs = experiment_desc["tasks"]
    params = experiment_desc.get("parameters", {})

    issues = []
    for step_name, step_def in graph.items():
        for ref in util.get_references(step_def):
            message = None

            dot_idx = ref.find(".")
            if dot_idx >= 0:
                ref_name = ref[:dot_idx]
                ref_output = ref[dot_idx + 1 :]
            else:
                ref_name = ref
                ref_output = None

            if ref_name in graph:
                referent_step = graph[ref_name]
                task_plugin_short_name = util.step_get_plugin_short_name(referent_step)
                # unrecognized task plugin short name is a different check.
                # We will disregard that possibility here.
                task_def = task_defs[task_plugin_short_name]
                task_outputs = task_def.get("outputs", [])

                if not isinstance(task_outputs, list):
                    task_outputs = [task_outputs]

                task_output_names = [
                    next(iter(task_output)) for task_output in task_outputs
                ]

                if ref_output is None:
                    if not task_outputs:
                        message = (
                            'In step "{}": reference "{}": referenced'
                            " step produces no output".format(step_name, ref)
                        )
                    elif len(task_outputs) > 1:
                        message = (
                            'In step "{}": reference "{}": an output name must'
                            " be given if the task plugin produces more than"
                            " one output."
                        ).format(step_name, ref)

                elif ref_output not in task_output_names:
                    message = (
                        'In step "{}": reference "{}": unrecognized output: {}'
                    ).format(step_name, ref, ref_output)

            elif ref_name in params:
                if ref_output:
                    message = (
                        'In step "{}": reference "{}": references to parameters'
                        " may not include an output name: {}"
                    ).format(step_name, ref, ref_output)

            else:
                message = 'In step "{}": unresolvable reference: {}'.format(
                    step_name, ref_name
                )

            if message:
                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )
                issues.append(issue)

    return issues


def _check_graph_dependencies(
    experiment_desc: Mapping[str, Any]
) -> list[ValidationIssue]:
    """
    Check explicitly declared dependencies for each step and ensure they refer
    to other steps.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    graph = experiment_desc["graph"]

    issues = []
    for step_name, step_def in graph.items():
        deps = step_def.get("dependencies", [])
        if isinstance(deps, str):
            deps = {deps}
        else:
            deps = set(deps)

        unrecognized_deps = deps - graph.keys()

        if unrecognized_deps:
            message = 'In step "{}": unrecognized dependency step(s): {}'.format(
                step_name, ", ".join(unrecognized_deps)
            )
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

    return issues


def _check_graph_cycle(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Check for a cycle in the task graph.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    issues = []
    message = None
    try:
        util.get_sorted_steps(experiment_desc["graph"])
    except BaseTaskEngineError as e:
        # If all references resolve and the description is schema-valid, the
        # only exception that could be thrown is a StepReferenceCycleError.
        # But I will catch all task engine errors, just in case.
        message = str(e)

    if message:
        issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
        issues.append(issue)

    return issues


def _check_names_dots(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Check whether any parameter or step names have a dot.  That needs to be
    disallowed because references to these would have the same syntax as a step
    output (<step>.<output>) and be ambiguous or misinterpreted.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """
    parameters = experiment_desc.get("parameters", [])
    graph = experiment_desc["graph"]

    issues = []
    for param_name in parameters:
        # The check for string keys is separate and prerequisite for this
        # check, so we will assume keys are strings here.
        if "." in param_name:
            message = (
                'Parameter name "{}" contains a dot.  References to this'
                " parameter will ambiguously look like step.output"
                " references."
            ).format(param_name)

            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)

            issues.append(issue)

    for step_name in graph:
        if "." in step_name:
            message = (
                'Step name "{}" contains a dot.  References to this'
                " step will ambiguously look like step.output"
                " references."
            ).format(step_name)

            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)

            issues.append(issue)

    return issues


def _check_step_structure(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Ensure each graph step includes a reference to a task plugin.  This check
    is about the structure of the step, not whether the reference resolves or
    whether the invocation makes sense with respect to the task plugin.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    # I tried writing JSON-Schema for this, but I felt it was too complicated
    # (if it had worked at all).  So I wrote some code to check instead.

    graph = experiment_desc["graph"]

    issues = []
    for step_name, step_def in graph.items():
        message = None
        if "task" not in step_def:
            # This is a short-form positional or keyword arg invocation.
            if len(step_def) == 1 and "dependencies" in step_def:
                # As of this writing, this structure does not pass schema
                # validation (i.e. an object with one property which is
                # "dependencies").  So this check won't get a chance to
                # succeed.  I'd like to keep this check here anyway, just in
                # case.
                message = (
                    'In step "{}": illegal task invocation: A'
                    " positional/keyword short form step definition must be an"
                    " object which includes a property whose name is the"
                    " task plugin short name."
                ).format(step_name)

            elif len(step_def) == 2 and "dependencies" not in step_def:
                message = (
                    'In step "{}": illegal task invocation: A'
                    " positional/keyword short form step definition must be an"
                    " object which includes a property whose name is the"
                    " task plugin short name."
                ).format(step_name)

            # Presence of more than two properties is checked in schema.

        if not message:
            # as another safety check if the above checks find no issues,
            # ensure we can get a task plugin short name from the step.
            task_short_name = util.step_get_plugin_short_name(step_def)
            if task_short_name is None:
                message = (
                    'In step "{}": illegal task invocation: unable to'
                    " determine task plugin short name"
                ).format(step_name)

        if message:
            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)
            issues.append(issue)

    return issues


def _check_task_invocation(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Check task invocation args against declared task inputs, with regard
    to name, number, etc (but not type).

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    tasks = experiment_desc["tasks"]
    graph = experiment_desc["graph"]

    issues = []

    for step_name, step_def in graph.items():
        task_short_name = util.step_get_plugin_short_name(step_def)
        task_def = tasks[task_short_name]

        task_input_map = util.make_task_input_map(task_def)
        invoc_pos_args, invoc_kwargs = util.step_get_invocation_arg_specs(step_def)

        # step_get_invocation_arg_specs() returns nulls if the step definition
        # is malformed, but we already checked that.  This code does not run
        # unless the step definition is well formed.
        assert invoc_pos_args is not None
        assert invoc_kwargs is not None

        # Sanity check number of positional args given
        num_pos_args = len(invoc_pos_args)
        num_task_inputs = len(task_input_map)
        if num_pos_args > num_task_inputs:
            message = (
                'In step "{}": illegal task invocation: there are more'
                " positional args given than task parameters: {} > {}"
            ).format(step_name, num_pos_args, num_task_inputs)

            issue = ValidationIssue(IssueType.SEMANTIC, IssueSeverity.ERROR, message)

            issues.append(issue)

        # Check for keyword args in the invocation, which do not correspond to
        # defined task parameters
        for invoc_kwarg_name in invoc_kwargs:
            if invoc_kwarg_name not in task_input_map:
                message = (
                    'In step "{}": illegal task invocation: keyword arg does'
                    " not correspond to a task parameter: {}"
                ).format(step_name, invoc_kwarg_name)

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

        # Check for task parameters assigned both positionally and as keyword.
        # Need to zip(invoc_pos_args, task_input_map) to limit the search
        # through task parameters to just those which have positional
        # assignments.  But invoc_pos_args is not otherwise used.  This causes
        # a flake8 error about an unused loop variable.  That error is quieted
        # by adding a leading underscore.
        for idx, (_invoc_pos_arg, task_param_name) in enumerate(
            zip(invoc_pos_args, task_input_map)
        ):
            if task_param_name in invoc_kwargs:
                message = (
                    'In step "{}": illegal task invocation: task parameter'
                    ' "{}" is assigned a value both as a keyword arg and'
                    " positional arg #{}"
                ).format(step_name, task_param_name, idx + 1)

                issue = ValidationIssue(
                    IssueType.SEMANTIC, IssueSeverity.ERROR, message
                )

                issues.append(issue)

        # Check for required parameters which are missing in the invocation
        for idx, (task_param_name, task_param_def) in enumerate(task_input_map.items()):
            required = True
            if "name" in task_param_def:
                required = task_param_def.get("required", True)

            if required:
                present = (
                    len(invoc_pos_args) >= idx + 1
                ) or task_param_name in invoc_kwargs

                if not present:
                    message = (
                        'In step "{}": illegal task invocation: required task'
                        " parameter is not assigned a value: {}"
                    ).format(step_name, task_param_name)

                    issue = ValidationIssue(
                        IssueType.SEMANTIC, IssueSeverity.ERROR, message
                    )

                    issues.append(issue)

    return issues


def _any_errors(issues: Iterable[ValidationIssue]) -> bool:
    """
    Check whether the given iterable of issues includes any of "error"
    severity.

    Args:
        issues: An iterable of validation issues

    Returns:
        True if there are any errors; False if not
    """
    return any(issue.severity is IssueSeverity.ERROR for issue in issues)


def _manually_validate(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Do any extra domain-specific handwritten validation we can think of, which
    can't be done (or awkward to do) via JSON-Schema.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    issues = []

    string_key_issues = _check_string_keys(experiment_desc)
    issues += string_key_issues

    if not string_key_issues:
        # Obviously the question of dots in names is moot if the names aren't
        # even strings!
        issues += _check_names_dots(experiment_desc)

    issues += _check_task_plugin_io_names(experiment_desc)
    issues += _check_task_plugin_io_types(experiment_desc)
    issues += _check_task_plugin_pyplugs_coords(experiment_desc)

    step_structure_issues = _check_step_structure(experiment_desc)
    issues += step_structure_issues

    name_collision_issues = _check_name_collisions(experiment_desc)
    issues += name_collision_issues

    if not step_structure_issues:
        task_ref_issues = _check_task_plugin_references(experiment_desc)
        issues += task_ref_issues

        # The below checks require correct references (via task short name)
        # to task definitions, so we can find their input requirements.
        if not task_ref_issues:
            issues += _check_task_invocation(experiment_desc)

            # If there were name collisions, we can't properly interpret
            # references.  So we will skip these checks.
            if not name_collision_issues:
                graph_ref_issues = _check_graph_references(experiment_desc)
                graph_ref_issues += _check_graph_dependencies(experiment_desc)
                issues += graph_ref_issues

                # The graph topology is based on references.  If there were
                # reference errors, we don't have sensible graph.  So we aren't
                # able to check the graph for cycles.  So we will skip this
                # check.
                if not graph_ref_issues:
                    issues += _check_graph_cycle(experiment_desc)

    issues += _check_global_parameter_types(experiment_desc)
    issues += _check_type_definition_type_references(experiment_desc)
    issues += _check_type_reference_cycle(experiment_desc)

    # We must have basic things like correct types, tasks, steps, global
    # parameters, and resolvable references therein, before type validation can
    # be expected to succeed.  So maybe make type validation depend on
    # everything above?
    if not _any_errors(issues):
        issues += type_validation.check_types(experiment_desc)

        # This check re-creates the type registry, which can error out if
        # there were type issues detected above (e.g. a type reference cycle).
        # So let's skip the check if there were other errors.
        issues += _check_union_member_duplicates(experiment_desc)

    return issues


def validate(experiment_desc: Mapping[str, Any]) -> list[ValidationIssue]:
    """
    Validate the given declarative experiment description.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        A list of ValidationIssue objects; will be an empty list if the
        experiment description was valid.
    """

    issues = _schema_validate(experiment_desc)

    # If the description is not schema-valid, the basic structure is incorrect,
    # so we won't even try to dig inside it to check anything.
    if not issues:
        issues = _manually_validate(experiment_desc)

    return issues


def is_valid(experiment_desc: Mapping[str, Any]) -> bool:
    """
    Validate the given declarative experiment description and return a simple
    boolean result.  This style is more appropriate in contexts where we only
    care about whether the description is valid or not, and don't need to know
    the reasons.

    Args:
        experiment_desc: The experiment description, as parsed YAML or
            equivalent

    Returns:
        True if the description is valid; False if not
    """

    issues = validate(experiment_desc)

    # If it's warnings only, let's treat that as valid.
    result = not _any_errors(issues)

    return result



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Unit tests for the type classes.
"""
import pytest

from dioptra.sdk.exceptions.task_engine import InvalidTypeStructureDefinitionError
from dioptra.task_engine.type_registry import TYPE_INTEGER, TYPE_STRING
from dioptra.task_engine.types import (
    SimpleType,
    StructuredType,
    StructureType,
    TypeStructure,
    UnionType,
)


def test_simple_type() -> None:
    super_type = SimpleType("A")
    sub_type = SimpleType("B", super_type=super_type)

    assert super_type.name == "A"
    assert sub_type.name == "B"
    assert sub_type.super_type == super_type
    assert sub_type.is_subtype_of(super_type)
    assert not sub_type.is_subtype_of(sub_type)
    assert not super_type.is_subtype_of(sub_type)

    with pytest.raises(ValueError):
        SimpleType(None)


def test_simple_type_eq() -> None:
    type1 = SimpleType("B")
    type2 = SimpleType("B")

    assert type1 == type2
    assert hash(type1) == hash(type2)

    super_type = SimpleType("A")

    type1 = SimpleType("B", super_type=super_type)
    type2 = SimpleType("B", super_type=super_type)

    assert type1 == type2
    assert hash(type1) == hash(type2)


def test_simple_type_neq() -> None:
    type1 = SimpleType("A")
    type2 = SimpleType("B")

    assert type1 != type2
    # For the not-equal tests, hashes may be equal or not equal; there are
    # no requirements which we can subject to testing.


def test_list_structure() -> None:
    elt_type = SimpleType("A")

    list_struct = TypeStructure(StructureType.LIST, elt_type)

    assert list_struct.struct_type is StructureType.LIST
    assert list_struct.struct_def == elt_type

    with pytest.raises(InvalidTypeStructureDefinitionError):
        TypeStructure(StructureType.LIST, [elt_type])

    with pytest.raises(InvalidTypeStructureDefinitionError):
        TypeStructure(StructureType.LIST, {"foo": elt_type})


def test_list_structure_eq() -> None:
    elt_type = SimpleType("A")

    list_struct1 = TypeStructure(StructureType.LIST, elt_type)

    list_struct2 = TypeStructure(StructureType.LIST, elt_type)

    assert list_struct1 == list_struct2
    assert hash(list_struct1) == hash(list_struct2)


def test_list_structure_neq() -> None:
    elt_type1 = SimpleType("A")
    elt_type2 = SimpleType("B")

    list_struct1 = TypeStructure(StructureType.LIST, elt_type1)

    list_struct2 = TypeStructure(StructureType.LIST, elt_type2)

    assert list_struct1 != list_struct2


def test_tuple_structure() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    tuple_struct = TypeStructure(StructureType.TUPLE, [a_type, b_type])

    assert tuple_struct.struct_type is StructureType.TUPLE
    assert tuple_struct.struct_def == [a_type, b_type]

    with pytest.raises(InvalidTypeStructureDefinitionError):
        TypeStructure(StructureType.TUPLE, a_type)

    with pytest.raises(InvalidTypeStructureDefinitionError):
        TypeStructure(StructureType.TUPLE, {"foo": a_type})

    # Allowed corner case: empty tuple structure
    TypeStructure(StructureType.TUPLE, [])


def test_tuple_structure_eq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    tuple_struct1 = TypeStructure(StructureType.TUPLE, [a_type, b_type])

    tuple_struct2 = TypeStructure(StructureType.TUPLE, [a_type, b_type])

    assert tuple_struct1 == tuple_struct2
    assert hash(tuple_struct1) == hash(tuple_struct2)


def test_tuple_structure_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    c_type = SimpleType("C")

    tuple_struct1 = TypeStructure(StructureType.TUPLE, [a_type, b_type])

    tuple_struct2 = TypeStructure(StructureType.TUPLE, [b_type, a_type])

    tuple_struct3 = TypeStructure(StructureType.TUPLE, [b_type, c_type])

    assert tuple_struct1 != tuple_struct2
    assert tuple_struct2 != tuple_struct3


def test_mapping_structure_enum() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    map_struct = TypeStructure(StructureType.MAPPING, {"foo": a_type, "bar": b_type})

    assert map_struct.struct_type is StructureType.MAPPING
    assert map_struct.struct_def == {"foo": a_type, "bar": b_type}

    with pytest.raises(InvalidTypeStructureDefinitionError):
        TypeStructure(StructureType.MAPPING, a_type)

    # Allowed corner case: empty enumerated mapping structure
    TypeStructure(StructureType.MAPPING, {})


def test_mapping_structure_enum_eq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    map_struct1 = TypeStructure(
        StructureType.MAPPING,
        dict(
            [
                # This construction allows us to force particular dict iteration
                # orders.
                ("foo", a_type),
                ("bar", b_type),
            ]
        ),
    )

    map_struct2 = TypeStructure(
        StructureType.MAPPING, dict([("bar", b_type), ("foo", a_type)])
    )

    assert map_struct1 == map_struct2
    assert hash(map_struct1) == hash(map_struct2)


def test_mapping_structure_enum_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    c_type = SimpleType("C")

    map_struct1 = TypeStructure(StructureType.MAPPING, {"foo": a_type, "bar": b_type})

    map_struct2 = TypeStructure(StructureType.MAPPING, {"bar": b_type, "foo": c_type})

    map_struct3 = TypeStructure(StructureType.MAPPING, {"baz": a_type, "bar": b_type})

    assert map_struct1 != map_struct2
    assert map_struct1 != map_struct3


def test_mapping_structure_key_value_type() -> None:
    a_type = SimpleType("A")

    map_struct = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])

    assert map_struct.struct_type is StructureType.MAPPING
    assert map_struct.struct_def == [TYPE_STRING, a_type]


def test_mapping_structure_key_value_eq() -> None:
    a_type = SimpleType("A")

    map_struct1 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])

    map_struct2 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])

    assert map_struct1 == map_struct2
    assert hash(map_struct1) == hash(map_struct2)


def test_mapping_structure_key_value_neq() -> None:
    a_type = SimpleType("A")

    map_struct1 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])

    map_struct2 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, TYPE_INTEGER])

    map_struct3 = TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, a_type])

    assert map_struct1 != map_struct2
    assert map_struct1 != map_struct3
    assert map_struct2 != map_struct3


def test_structured_type_list() -> None:
    elt_type = SimpleType("A")
    list_structure = TypeStructure(StructureType.LIST, elt_type)
    list_type = StructuredType(list_structure, name="list_of_A")

    assert list_type.name == "list_of_A"
    assert list_type.structure == list_structure

    list_type_anon = StructuredType(list_structure)

    assert not list_type_anon.name
    assert list_type_anon.structure == list_structure


def test_structured_type_list_eq() -> None:
    elt_type = SimpleType("A")
    list_structure = TypeStructure(StructureType.LIST, elt_type)
    list_type1 = StructuredType(list_structure, name="list_of_A")
    list_type2 = StructuredType(list_structure, name="list_of_A")
    list_type_anon1 = StructuredType(list_structure)
    list_type_anon2 = StructuredType(list_structure)

    assert list_type1 == list_type2
    assert hash(list_type1) == hash(list_type2)

    assert list_type_anon1 == list_type_anon2
    assert hash(list_type_anon1) == hash(list_type_anon2)


def test_structured_type_list_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    list_structure1 = TypeStructure(StructureType.LIST, a_type)
    list_structure2 = TypeStructure(StructureType.LIST, b_type)
    list_type1 = StructuredType(list_structure1, name="list_of_A")
    list_type2 = StructuredType(list_structure2, name="list_of_B")
    list_type_anon1 = StructuredType(list_structure1)
    list_type_anon2 = StructuredType(list_structure2)

    assert list_type1 != list_type2
    assert list_type1 != list_type_anon1
    assert list_type_anon1 != list_type_anon2


def test_structured_type_tuple() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    tuple_structure = TypeStructure(StructureType.TUPLE, [a_type, b_type])
    tuple_type = StructuredType(tuple_structure, name="tuple_A_B")
    tuple_type_anon = StructuredType(tuple_structure)

    assert tuple_type.name == "tuple_A_B"
    assert tuple_type.structure == tuple_structure

    assert not tuple_type_anon.name
    assert tuple_type_anon.structure == tuple_structure

    # Allowed corner case: empty tuple type
    StructuredType(TypeStructure(StructureType.TUPLE, []), name="empty_tuple")


def test_structured_type_tuple_eq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    tuple_structure = TypeStructure(StructureType.TUPLE, [a_type, b_type])
    tuple_type1 = StructuredType(tuple_structure, name="tuple_A_B")
    tuple_type2 = StructuredType(tuple_structure, name="tuple_A_B")
    tuple_type_anon1 = StructuredType(tuple_structure)
    tuple_type_anon2 = StructuredType(tuple_structure)

    assert tuple_type1 == tuple_type2
    assert hash(tuple_type1) == hash(tuple_type2)

    assert tuple_type_anon1 == tuple_type_anon2
    assert hash(tuple_type_anon1) == hash(tuple_type_anon2)


def test_structured_type_tuple_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    c_type = SimpleType("C")
    tuple_structure1 = TypeStructure(StructureType.TUPLE, [a_type, b_type])
    tuple_structure2 = TypeStructure(StructureType.TUPLE, [b_type, a_type])
    tuple_structure3 = TypeStructure(StructureType.TUPLE, [b_type, c_type])
    tuple_type1 = StructuredType(tuple_structure1, name="tuple_A_B")
    tuple_type2 = StructuredType(tuple_structure2, name="tuple_B_A")
    tuple_type3 = StructuredType(tuple_structure3, name="tuple_B_C")
    tuple_type_anon1 = StructuredType(tuple_structure1)
    tuple_type_anon2 = StructuredType(tuple_structure2)
    tuple_type_anon3 = StructuredType(tuple_structure2)

    assert tuple_type1 != tuple_type2
    assert tuple_type1 != tuple_type3
    assert tuple_type1 != tuple_type_anon1
    assert tuple_type1 != tuple_type_anon2
    assert tuple_type1 != tuple_type_anon3
    assert tuple_type_anon1 != tuple_type_anon2
    assert tuple_type_anon1 != tuple_type_anon3


def test_structured_type_mapping_enum() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    map_structure = TypeStructure(StructureType.MAPPING, {"foo": a_type, "bar": b_type})
    map_type = StructuredType(map_structure, name="map_foo_bar")
    map_type_anon = StructuredType(map_structure)

    assert map_type.name == "map_foo_bar"
    assert map_type.structure == map_structure
    assert not map_type_anon.name
    assert map_type_anon.structure == map_structure

    # Allowed corner case: empty enumerated mapping type
    StructuredType(TypeStructure(StructureType.MAPPING, {}), name="empty_enum_mapping")


def test_structured_type_mapping_enum_eq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    map_structure1 = TypeStructure(
        StructureType.MAPPING, dict([("foo", a_type), ("bar", b_type)])
    )
    map_structure2 = TypeStructure(
        StructureType.MAPPING, dict([("bar", b_type), ("foo", a_type)])
    )
    map_type1 = StructuredType(map_structure1, name="map_foo_bar")
    map_type2 = StructuredType(map_structure2, name="map_foo_bar")
    map_type_anon1 = StructuredType(map_structure1)
    map_type_anon2 = StructuredType(map_structure2)

    assert map_type1 == map_type2
    assert hash(map_type1) == hash(map_type2)

    assert map_type_anon1 == map_type_anon2
    assert hash(map_type_anon1) == hash(map_type_anon2)


def test_structured_type_mapping_enum_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    c_type = SimpleType("C")
    map_structure1 = TypeStructure(
        StructureType.MAPPING, dict([("foo", a_type), ("bar", b_type)])
    )
    map_structure2 = TypeStructure(
        StructureType.MAPPING, dict([("bar", c_type), ("foo", a_type)])
    )
    map_structure3 = TypeStructure(
        StructureType.MAPPING, dict([("foo", a_type), ("baz", b_type)])
    )
    map_type1 = StructuredType(map_structure1, name="map_foo_bar")
    map_type2 = StructuredType(map_structure2, name="map_bar_foo")
    map_type3 = StructuredType(map_structure3, name="map_foo_baz")
    map_type_anon1 = StructuredType(map_structure1)
    map_type_anon2 = StructuredType(map_structure2)
    map_type_anon3 = StructuredType(map_structure3)

    assert map_type1 != map_type2
    assert map_type1 != map_type3
    assert map_type1 != map_type_anon1
    assert map_type1 != map_type_anon2
    assert map_type1 != map_type_anon3
    assert map_type_anon1 != map_type_anon2
    assert map_type_anon1 != map_type_anon3


def test_structured_type_mapping_key_value() -> None:
    a_type = SimpleType("A")
    map_structure = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    map_type = StructuredType(map_structure, name="map_string_a")
    map_type_anon = StructuredType(map_structure)

    assert map_type.name == "map_string_a"
    assert map_type.structure == map_structure

    assert not map_type_anon.name
    assert map_type_anon.structure == map_structure


def test_structured_type_mapping_key_value_eq() -> None:
    a_type = SimpleType("A")
    map_structure = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    map_type1 = StructuredType(map_structure, name="map_string_a")
    map_type2 = StructuredType(map_structure, name="map_string_a")
    map_type_anon1 = StructuredType(map_structure)
    map_type_anon2 = StructuredType(map_structure)

    assert map_type1 == map_type2
    assert hash(map_type1) == hash(map_type2)

    assert map_type_anon1 == map_type_anon2
    assert hash(map_type_anon1) == hash(map_type_anon2)


def test_structured_type_mapping_key_value_neq() -> None:
    a_type = SimpleType("A")
    map_structure1 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, a_type])
    map_structure2 = TypeStructure(StructureType.MAPPING, [TYPE_STRING, TYPE_INTEGER])
    map_structure3 = TypeStructure(StructureType.MAPPING, [TYPE_INTEGER, a_type])
    map_type1 = StructuredType(map_structure1, name="map_string_a")
    map_type2 = StructuredType(map_structure2, name="map_string_int")
    map_type3 = StructuredType(map_structure3, name="map_int_a")
    map_type_anon1 = StructuredType(map_structure1)
    map_type_anon2 = StructuredType(map_structure2)
    map_type_anon3 = StructuredType(map_structure3)

    assert map_type1 != map_type2
    assert map_type1 != map_type3
    assert map_type1 != map_type_anon1
    assert map_type1 != map_type_anon2
    assert map_type1 != map_type_anon3
    assert map_type2 != map_type3
    assert map_type_anon1 != map_type_anon2
    assert map_type_anon1 != map_type_anon3
    assert map_type_anon2 != map_type_anon3


def test_structured_type_no_structure() -> None:
    with pytest.raises(ValueError):
        StructuredType(None)


def test_union_type() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    union_type = UnionType([a_type, b_type], name="union_a_b")

    assert union_type.name == "union_a_b"
    assert union_type.member_types == {b_type, a_type}

    union_type_anon = UnionType([a_type, b_type])

    assert not union_type_anon.name
    assert union_type_anon.member_types == {b_type, a_type}

    # Allowed corner case: empty union
    UnionType([], name="empty_union")


def test_union_type_eq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")

    union_type1 = UnionType([a_type, b_type], name="union_a_b")
    union_type2 = UnionType([b_type, a_type], name="union_a_b")
    union_type_anon1 = UnionType([a_type, b_type])
    union_type_anon2 = UnionType([b_type, a_type])

    assert union_type1 == union_type2
    assert hash(union_type1) == hash(union_type2)

    assert union_type_anon1 == union_type_anon2
    assert hash(union_type_anon1) == hash(union_type_anon2)


def test_union_type_neq() -> None:
    a_type = SimpleType("A")
    b_type = SimpleType("B")
    c_type = SimpleType("C")

    union_type1 = UnionType([a_type, b_type], name="union_a_b")
    union_type2 = UnionType([a_type, c_type], name="union_a_c")
    union_type_anon1 = UnionType([a_type, b_type])
    union_type_anon2 = UnionType([a_type, c_type])

    assert union_type1 != union_type2
    assert union_type1 != union_type_anon1
    assert union_type1 != union_type_anon2
    assert union_type_anon1 != union_type_anon2



"""Update uniqueness constraint name for users.alternative_id

Revision ID: 8ffe4640e009
Revises: ef1936dcc5eb
Create Date: 2024-02-09 18:23:45.140756

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "8ffe4640e009"
down_revision = "ef1936dcc5eb"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("legacy_users", schema=None) as batch_op:
        batch_op.drop_constraint("uq_users_alternative_id", type_="unique")
        batch_op.create_unique_constraint(
            batch_op.f("uq_legacy_users_alternative_id"), ["alternative_id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("legacy_users", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("uq_legacy_users_alternative_id"), type_="unique"
        )
        batch_op.create_unique_constraint("uq_users_alternative_id", ["alternative_id"])

    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform tag endpoint operations."""
from __future__ import annotations

import datetime
from typing import Any, Final

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    SortParameterValidationError,
)
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "tag"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.Tag.name.like(x),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.Tag.name,
    "createdOn": models.Tag.created_on,
    "lastModifiedOn": models.Tag.last_modified_on,
}


class TagService(object):
    """The service methods for registering and managing tags by their unique id."""

    @inject
    def __init__(
        self,
        tag_name_service: TagNameService,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the tag service.

        All arguments are provided via dependency injection.

        Args:
            tag_name_service: A TagNameService object.
            group_id_service: A GroupIdService object.
        """
        self._tag_name_service = tag_name_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        group_id: int,
        commit: bool = True,
        **kwargs,
    ) -> models.Tag:
        """Create a new tag.

        Args:
            name: The name of the tag. The combination of name and group_id must be
                unique.
            description: The description of the tag.
            group_id: The group that will own the tag.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created tag object.

        Raises:
            EntityExistsError: If a tag with the given name already exists.
            EntityDoesNotExistError: If the group with the provided ID does not exist.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        duplicate = self._tag_name_service.get(name, group_id=group_id, log=log)
        if duplicate is not None:
            raise EntityExistsError(
                RESOURCE_TYPE, duplicate.tag_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)

        new_tag = models.Tag(name=name, owner=group, creator=current_user)
        db.session.add(new_tag)

        if commit:
            db.session.commit()
            log.debug(
                "Tag creation successful", tag_id=new_tag.tag_id, name=new_tag.name
            )

        return new_tag

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[models.Tag], int]:
        """Fetch a list of tags, optionally filtering by search string and paging
        parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of tags to be returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of tags and the total number of tags matching
            the query.

        Raises:
            BackendDatabaseError: If the database query returns a None when counting
                the number of tags.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of tags")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        count_tags_stmt = select(func.count(models.Tag.tag_id)).where(*filters)
        total_num_tags = db.session.scalar(count_tags_stmt)

        if total_num_tags is None:
            log.error(
                "The database query returned a None when counting the number of "
                "tags when it should return a number.",
                sql=str(count_tags_stmt),
            )
            raise BackendDatabaseError

        tags_stmt = (
            select(models.Tag).where(*filters).offset(page_index).limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            tags_stmt = tags_stmt.order_by(sort_column)
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        tags = list(db.session.scalars(tags_stmt).all())

        return tags, total_num_tags


class TagIdService(object):
    """The service methods for registering and managing tags by their unique id."""

    @inject
    def __init__(
        self,
        tag_name_service: TagNameService,
    ) -> None:
        """Initialize the tag id service.

        All arguments are provided via dependency injection.

        Args:
            tag_name_service: A TagNameService object.
        """
        self._tag_name_service = tag_name_service

    def get(
        self,
        tag_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Tag | None:
        """Fetch a tag by its unique id.

        Args:
            tag_id: The unique id of the tag.
            error_if_not_found: If True, raise an error if the tag is not found.
                Defaults to False.

        Returns:
            The tag object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the tag is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get tag by id", tag_id=tag_id)

        stmt = select(models.Tag).where(models.Tag.tag_id == tag_id)
        tag = db.session.scalar(stmt)

        if tag is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, tag_id=tag_id)

            return None

        return tag

    def modify(
        self,
        tag_id: int,
        name: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> models.Tag | None:
        """Modify a tag.

        Args:
            tag_id: The unique id of the tag.
            name: The new name of the tag.
            description: The new description of the tag.
            error_if_not_found: If True, raise an error if the group is not found.
                Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated tag object.

        Raises:
            EntityDoesNotExistError: If the tag is not found and `error_if_not_found`
                is True.
            EntityExistsError: If the tag name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        tag = self.get(tag_id, error_if_not_found=error_if_not_found, log=log)

        if tag is None:
            return None

        group_id = tag.group_id
        if name != tag.name:
            duplicate = self._tag_name_service.get(name, group_id=group_id, log=log)
            if duplicate is not None:
                raise EntityExistsError(
                    RESOURCE_TYPE, duplicate.tag_id, name=name, group_id=group_id
                )

        current_timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        tag.name = name
        tag.last_modified_on = current_timestamp

        if commit:
            db.session.commit()
            log.debug(
                "Tag modification successful",
                tag_id=tag_id,
                name=name,
            )

        return tag

    def delete(
        self,
        tag_id: int,
        **kwargs,
    ) -> dict[str, Any]:
        """Delete a tag by its unique id.

        Args:
            tag_id: The unique id of the tag.

        Returns:
            The tag object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the tag is not found.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get tag by id", tag_id=tag_id)

        stmt = select(models.Tag).where(models.Tag.tag_id == tag_id)
        tag = db.session.scalar(stmt)

        if tag is None:
            raise EntityDoesNotExistError(RESOURCE_TYPE, tag_id=tag_id)

        tag_id = tag.tag_id
        db.session.delete(tag)
        db.session.commit()
        log.debug("Tag deleted", tag_id=tag_id)

        return {"status": "Success", "id": [tag_id]}


class TagIdResourcesService(object):
    """The service methods for retrieving resources with a given Tag."""

    @inject
    def __init__(
        self,
        tag_name_service: TagNameService,
    ) -> None:
        """Initialize the tag id service.

        All arguments are provided via dependency injection.

        Args:
            tag_name_service: A TagNameService object.
        """
        self._tag_name_service = tag_name_service

    def get(
        self,
        tag_id: int,
        resource_type: str | None,
        page_index: int,
        page_length: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> tuple[list[models.Resource], int] | None:
        """Fetch a list of resources with the Tag.

        Args:
            tag_id: The unique id of the tag.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of tags to be returned.
            error_if_not_found: If True, raise an error if the tag is not found.
                Defaults to False.

        Returns:
            The tag object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the tag is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get tag by id", tag_id=tag_id)

        tag_stmt = select(models.Tag).where(models.Tag.tag_id == tag_id)
        tag = db.session.scalar(tag_stmt)

        if tag is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(RESOURCE_TYPE, tag_id=tag_id)

            return None

        filters = []
        if resource_type is not None:
            filters.append(models.Resource.resource_type == resource_type)

        count_resources_stmt = select(func.count(models.Resource.resource_id)).where(
            *filters,
            models.Resource.tags.any(models.Tag.tag_id == tag_id),
        )
        total_num_resources = db.session.scalar(count_resources_stmt)

        if total_num_resources is None:
            log.error(
                "The database query returned a None when counting the number of "
                "resources when it should return a number.",
                sql=str(count_resources_stmt),
            )
            raise BackendDatabaseError

        resources_stmt = (
            select(models.Resource)
            .where(
                *filters,
                models.Resource.tags.any(models.Tag.tag_id == tag_id),
            )
            .offset(page_index)
            .limit(page_length)
        )
        resources = list(db.session.scalars(resources_stmt).all())

        return resources, total_num_resources


class TagNameService(object):
    """The service methods for managing tags by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.Tag | None:
        """Fetch a tag by its name.

        Args:
            name: The name of the tag.
            group_id: The the group id of the tag.
            error_if_not_found: If True, raise an error if the tag is not found.
                Defaults to False.

        Returns:
            The tag object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the tag is not found and `error_if_not_found`
                is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get tag by name", tag_name=name, group_id=group_id)

        stmt = select(models.Tag).where(
            models.Tag.name == name,
            models.Tag.group_id == group_id,
        )
        tag = db.session.scalars(stmt).first()

        if tag is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, name=name, group_id=group_id
                )

            return None

        return tag



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from .artifacts import Artifact
from .entry_points import (
    EntryPoint,
    EntryPointParameter,
    EntryPointParameterValue,
    EntryPointPluginFile,
    entry_point_parameter_types_table,
)
from .experiments import Experiment
from .groups import Group, GroupManager, GroupMember
from .jobs import (
    EntryPointJob,
    ExperimentJob,
    Job,
    JobMlflowRun,
    QueueJob,
    job_status_types_table,
)
from .locks import (
    GroupLock,
    ResourceLock,
    UserLock,
    group_lock_types_table,
    resource_lock_types_table,
    user_lock_types_table,
)
from .ml_models import MlModel, MlModelVersion
from .plugins import (
    Plugin,
    PluginFile,
    PluginTask,
    PluginTaskInputParameter,
    PluginTaskOutputParameter,
    PluginTaskParameterType,
)
from .queues import Queue
from .resources import (
    DraftResource,
    Resource,
    ResourceSnapshot,
    SharedResource,
    resource_dependencies_table,
    resource_dependency_types_table,
    resource_tags_table,
    resource_types_table,
    shared_resource_tags_table,
)
from .tags import Tag
from .users import User

__all__ = [
    "Artifact",
    "DraftResource",
    "EntryPoint",
    "EntryPointJob",
    "EntryPointParameter",
    "EntryPointParameterValue",
    "EntryPointPluginFile",
    "Experiment",
    "ExperimentJob",
    "Group",
    "GroupLock",
    "GroupManager",
    "GroupMember",
    "Job",
    "JobMlflowRun",
    "MlModel",
    "MlModelVersion",
    "Plugin",
    "PluginFile",
    "PluginTask",
    "PluginTaskInputParameter",
    "PluginTaskOutputParameter",
    "PluginTaskParameterType",
    "Queue",
    "QueueJob",
    "Resource",
    "ResourceLock",
    "ResourceSnapshot",
    "SharedResource",
    "Tag",
    "User",
    "UserLock",
    "entry_point_parameter_types_table",
    "group_lock_types_table",
    "job_status_types_table",
    "resource_dependencies_table",
    "resource_dependency_types_table",
    "resource_lock_types_table",
    "resource_tags_table",
    "resource_types_table",
    "shared_resource_tags_table",
    "user_lock_types_table",
]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The server-side functions that perform plugin parameter type endpoint operations."""
from typing import Any, Final

import structlog
from flask_login import current_user
from injector import inject
from sqlalchemy import Integer, func, select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.db.models.constants import resource_lock_types
from dioptra.restapi.errors import (
    BackendDatabaseError,
    EntityDoesNotExistError,
    EntityExistsError,
    PluginParameterTypeMatchesBuiltinTypeError,
    ReadOnlyLockError,
    SortParameterValidationError,
)
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.groups.service import GroupIdService
from dioptra.restapi.v1.shared.search_parser import construct_sql_query_filters
from dioptra.task_engine.type_registry import BUILTIN_TYPES

LOGGER: BoundLogger = structlog.stdlib.get_logger()

RESOURCE_TYPE: Final[str] = "plugin_task_parameter_type"
SEARCHABLE_FIELDS: Final[dict[str, Any]] = {
    "name": lambda x: models.PluginTaskParameterType.name.like(x, escape="/"),
    "description": lambda x: models.PluginTaskParameterType.description.like(
        x, escape="/"
    ),
    "tag": lambda x: models.PluginTaskParameterType.tags.any(
        models.Tag.name.like(x, escape="/")
    ),
}
SORTABLE_FIELDS: Final[dict[str, Any]] = {
    "name": models.PluginTaskParameterType.name,
    "createdOn": models.PluginTaskParameterType.created_on,
    "lastModifiedOn": models.Resource.last_modified_on,
    "description": models.PluginTaskParameterType.description,
}


class PluginParameterTypeService(object):
    """The service methods for registering and managing plugin parameter types
    by their unique id."""

    @inject
    def __init__(
        self,
        plugin_parameter_type_name_service: "PluginParameterTypeNameService",
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the plugin parameter type service.

        All arguments are provided via dependency injection.

        Args:
            plugin_parameter_type_name_service: A PluginParameterTypeNameService object.
            group_id_service: A GroupIdService object.
        """
        self._plugin_parameter_type_name_service = plugin_parameter_type_name_service
        self._group_id_service = group_id_service

    def create(
        self,
        name: str,
        structure: dict[str, Any],
        description: str,
        group_id: int,
        commit: bool = True,
        **kwargs,
    ) -> utils.PluginParameterTypeDict:
        """Create a new plugin parameter type.

        Args:
            name: The name of the plugin parameter type. The combination of
                name and group_id must be unique. The name also cannot match
                any built-in types.
            structure: Optional JSON-type field for further constraining a
                type's structure.
            description: The description of the plugin parameter type.
            group_id: The group that will own the plugin parameter type.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The newly created plugin parameter type object.

        Raises:
            PluginParameterTypeMatchesBuiltinTypeError: If the plugin parameter
                type name matches a built-in type.
            EntityExistsError: If a plugin parameter type with the given name already
                exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        if name.strip().lower() in BUILTIN_TYPES:
            log.debug(
                "Plugin Parameter Type name matches a built-in type",
                name=name,
                normalized_name=name.strip().lower(),
            )
            raise PluginParameterTypeMatchesBuiltinTypeError

        duplicate = self._plugin_parameter_type_name_service.get(
            name, group_id=group_id, log=log
        )
        if duplicate is not None:
            raise EntityExistsError(
                RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
            )

        group = self._group_id_service.get(group_id, error_if_not_found=True)

        resource = models.Resource(resource_type=RESOURCE_TYPE, owner=group)
        new_plugin_parameter_type = models.PluginTaskParameterType(
            name=name,
            structure=structure,
            description=description,
            resource=resource,
            creator=current_user,
        )
        db.session.add(new_plugin_parameter_type)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin Parameter Type registration successful",
                plugin_parameter_type_id=new_plugin_parameter_type.resource_id,
                name=new_plugin_parameter_type.name,
            )

        return utils.PluginParameterTypeDict(
            plugin_task_parameter_type=new_plugin_parameter_type, has_draft=False
        )

    def get(
        self,
        group_id: int | None,
        search_string: str,
        page_index: int,
        page_length: int,
        sort_by_string: str,
        descending: bool,
        **kwargs,
    ) -> tuple[list[utils.PluginParameterTypeDict], int]:
        """Fetch a list of plugin parameter types, optionally filtering by
        search string and paging parameters.

        Args:
            group_id: A group ID used to filter results.
            search_string: A search string used to filter results.
            page_index: The index of the first group to be returned.
            page_length: The maximum number of plugin parameter types to be
                returned.
            sort_by_string: The name of the column to sort.
            descending: Boolean indicating whether to sort by descending or not.

        Returns:
            A tuple containing a list of plugin parameter types and the total
            number of plugin parameter types matching the query.

        Raises:
            SearchNotImplementedError: If a search string is provided.
            BackendDatabaseError: If the database query returns a None when
                counting the number of plugin parameter types.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug("Get full list of plugin parameter types")

        filters = list()

        if group_id is not None:
            filters.append(models.Resource.group_id == group_id)

        if search_string:
            filters.append(
                construct_sql_query_filters(search_string, SEARCHABLE_FIELDS)
            )

        stmt = (
            select(func.count(models.PluginTaskParameterType.resource_id))
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginTaskParameterType.resource_snapshot_id,
            )
        )
        total_num_plugin_parameter_types = db.session.scalars(stmt).first()

        if total_num_plugin_parameter_types is None:
            log.error(
                "The database query returned a None when counting the number of "
                "groups when it should return a number.",
                sql=str(stmt),
            )
            raise BackendDatabaseError

        if total_num_plugin_parameter_types == 0:
            return [], total_num_plugin_parameter_types

        plugin_parameter_types_stmt = (
            select(models.PluginTaskParameterType)
            .join(models.Resource)
            .where(
                *filters,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginTaskParameterType.resource_snapshot_id,
            )
            .offset(page_index)
            .limit(page_length)
        )

        if sort_by_string and sort_by_string in SORTABLE_FIELDS:
            sort_column = SORTABLE_FIELDS[sort_by_string]
            if descending:
                sort_column = sort_column.desc()
            else:
                sort_column = sort_column.asc()
            plugin_parameter_types_stmt = plugin_parameter_types_stmt.order_by(
                sort_column
            )
        elif sort_by_string and sort_by_string not in SORTABLE_FIELDS:
            raise SortParameterValidationError(RESOURCE_TYPE, sort_by_string)

        plugin_parameter_types = list(
            db.session.scalars(plugin_parameter_types_stmt).all()
        )

        drafts_stmt = select(
            models.DraftResource.payload["resource_id"].as_string().cast(Integer)
        ).where(
            models.DraftResource.payload["resource_id"]
            .as_string()
            .cast(Integer)
            .in_(
                tuple(
                    plugin_parameter_type.resource_id
                    for plugin_parameter_type in plugin_parameter_types
                )
            ),
            models.DraftResource.user_id == current_user.user_id,
        )
        plugin_parameter_types_dict: dict[int, utils.PluginParameterTypeDict] = {
            plugin_parameter_type.resource_id: utils.PluginParameterTypeDict(
                plugin_task_parameter_type=plugin_parameter_type, has_draft=False
            )
            for plugin_parameter_type in plugin_parameter_types
        }
        for resource_id in db.session.scalars(drafts_stmt):
            plugin_parameter_types_dict[resource_id]["has_draft"] = True

        return (
            list(plugin_parameter_types_dict.values()),
            total_num_plugin_parameter_types,
        )


class PluginParameterTypeIdService(object):
    """The service methods for registering and managing plugin parameter type
    by their unique id."""

    @inject
    def __init__(
        self,
        plugin_parameter_type_name_service: "PluginParameterTypeNameService",
    ) -> None:
        """Initialize the plugin parameter type service.

        All arguments are provided via dependency injection.

        Args:
            plugin_parameter_type_name_service: A PluginParameterTypeNameService object.
        """
        self._plugin_parameter_type_name_service = plugin_parameter_type_name_service

    def get(
        self,
        plugin_parameter_type_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> utils.PluginParameterTypeDict | None:
        """Fetch a plugin parameter type by its unique id.

        Args:
            plugin_parameter_type_id: The unique id of the plugin parameter
                type.
            error_if_not_found: If True, raise an error if the plugin parameter
                type is not found. Defaults to False.

        Returns:
            The plugin parameter type object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin parameter type
                is not found and `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get plugin parameter type by id",
            plugin_parameter_type_id=plugin_parameter_type_id,
        )

        stmt = (
            select(models.PluginTaskParameterType)
            .join(models.Resource)
            .where(
                models.PluginTaskParameterType.resource_id == plugin_parameter_type_id,
                models.PluginTaskParameterType.resource_snapshot_id
                == models.Resource.latest_snapshot_id,
                models.Resource.is_deleted == False,  # noqa: E712
            )
        )
        plugin_parameter_type = db.session.scalars(stmt).first()

        if plugin_parameter_type is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, plugin_parameter_type_id=plugin_parameter_type_id
                )

            return None

        drafts_stmt = (
            select(models.DraftResource.draft_resource_id)
            .where(
                models.DraftResource.payload["resource_id"].as_string().cast(Integer)
                == plugin_parameter_type.resource_id,
                models.DraftResource.user_id == current_user.user_id,
            )
            .exists()
            .select()
        )
        has_draft = db.session.scalar(drafts_stmt)

        return utils.PluginParameterTypeDict(
            plugin_task_parameter_type=plugin_parameter_type, has_draft=has_draft
        )

    def modify(
        self,
        plugin_parameter_type_id: int,
        name: str,
        structure: str,
        description: str,
        error_if_not_found: bool = False,
        commit: bool = True,
        **kwargs,
    ) -> utils.PluginParameterTypeDict | None:
        """Modify a plugin parameter type.

        Args:
            plugin_parameter_type_id: The unique id of the plugin parameter
                type.
            name: The new name of the plugin parameter type.
            structure: Optional JSON-type field for further constraining a
                type's structure.
            description: The new description of the plugin parameter type.
            error_if_not_found: If True, raise an error if the group is not
                found. Defaults to False.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            The updated plugin parameter type object.

        Raises:
            EntityDoesNotExistError: If the plugin parameter type
                is not found and `error_if_not_found` is True.
            EntityExistsError: If the plugin parameter type
                name already exists.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        plugin_parameter_type_dict = self.get(
            plugin_parameter_type_id, error_if_not_found=error_if_not_found, log=log
        )

        if plugin_parameter_type_dict is None:
            return None

        plugin_parameter_type = plugin_parameter_type_dict["plugin_task_parameter_type"]
        group_id = plugin_parameter_type.resource.group_id

        if plugin_parameter_type.resource.is_readonly:
            raise ReadOnlyLockError(
                RESOURCE_TYPE,
                plugin_parameter_type_id=plugin_parameter_type_id,
                name=plugin_parameter_type.name,
            )

        if name.strip().lower() in BUILTIN_TYPES:
            log.debug(
                "Plugin Parameter Type name matches a built-in type",
                name=name,
                normalized_name=name.strip().lower(),
            )
            raise PluginParameterTypeMatchesBuiltinTypeError

        if name != plugin_parameter_type.name:
            duplicate = self._plugin_parameter_type_name_service.get(
                name, group_id=group_id, log=log
            )
            if duplicate is not None:
                raise EntityExistsError(
                    RESOURCE_TYPE, duplicate.resource_id, name=name, group_id=group_id
                )

        new_plugin_parameter_type = models.PluginTaskParameterType(
            name=name,
            structure=structure,
            description=description,
            resource=plugin_parameter_type.resource,
            creator=current_user,
        )
        db.session.add(new_plugin_parameter_type)

        if commit:
            db.session.commit()
            log.debug(
                "Plugin Parameter Type modification successful",
                plugin_parameter_type_id=plugin_parameter_type_id,
                name=name,
                structure=structure,
                description=description,
            )

        return utils.PluginParameterTypeDict(
            plugin_task_parameter_type=new_plugin_parameter_type, has_draft=False
        )

    def delete(self, plugin_parameter_type_id: int, **kwargs) -> dict[str, Any]:
        """Delete a plugin parameter type.

        Args:
            plugin_parameter_type_id: The unique id of the plugin parameter type.

        Returns:
            A dictionary reporting the status of the request.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())

        stmt = select(models.Resource).filter_by(
            resource_id=plugin_parameter_type_id,
            resource_type=RESOURCE_TYPE,
            is_deleted=False,
        )
        plugin_parameter_type_resource = db.session.scalars(stmt).first()

        if plugin_parameter_type_resource is None:
            raise EntityDoesNotExistError(
                RESOURCE_TYPE, plugin_parameter_type_id=plugin_parameter_type_id
            )

        if plugin_parameter_type_resource.is_readonly:
            raise ReadOnlyLockError(
                RESOURCE_TYPE, plugin_parameter_type_id=plugin_parameter_type_id
            )

        deleted_resource_lock = models.ResourceLock(
            resource_lock_type=resource_lock_types.DELETE,
            resource=plugin_parameter_type_resource,
        )
        db.session.add(deleted_resource_lock)
        db.session.commit()
        log.debug(
            "Plugin Parameter Type deleted",
            plugin_parameter_type_id=plugin_parameter_type_id,
        )

        return {
            "status": "Success",
            "id": [plugin_parameter_type_id],
        }


class PluginParameterTypeNameService(object):
    """The service methods for managing plugin parameter types by their name."""

    def get(
        self,
        name: str,
        group_id: int,
        error_if_not_found: bool = False,
        **kwargs,
    ) -> models.PluginTaskParameterType | None:
        """Fetch a plugin parameter type by its name.

        Args:
            name: The name of the plugin parameter type.
            group_id: The the group id of the plugin parameter type.
            error_if_not_found: If True, raise an error if the plugin parameter
                type is not found. Defaults to False.

        Returns:
            The plugin parameter type object if found, otherwise None.

        Raises:
            EntityDoesNotExistError: If the plugin parameter type
                is not found and `error_if_not_found` is True.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        log.debug(
            "Get plugin parameter type by name",
            plugin_parameter_type_name=name,
            group_id=group_id,
        )

        stmt = (
            select(models.PluginTaskParameterType)
            .join(models.Resource)
            .where(
                models.PluginTaskParameterType.name == name,
                models.Resource.group_id == group_id,
                models.Resource.is_deleted == False,  # noqa: E712
                models.Resource.latest_snapshot_id
                == models.PluginTaskParameterType.resource_snapshot_id,
            )
        )
        plugin_parameter_type = db.session.scalars(stmt).first()

        if plugin_parameter_type is None:
            if error_if_not_found:
                raise EntityDoesNotExistError(
                    RESOURCE_TYPE, name=name, group_id=group_id
                )

            return None

        return plugin_parameter_type


class BuiltinPluginParameterTypeService(object):
    """The service methods for registering the built-in plugin parameter types"""

    @inject
    def __init__(
        self,
        group_id_service: GroupIdService,
    ) -> None:
        """Initialize the builtin plugin parameter type service.

        All arguments are provided via dependency injection.

        Args:
            group_id_service: A GroupIdService object.
        """
        self._group_id_service = group_id_service

    def create_all(
        self,
        user: models.User,
        group: models.Group,
        commit: bool = True,
        **kwargs,
    ) -> list[utils.PluginParameterTypeDict]:
        """Create all built-in plugin parameter types in a given group.

        Args:
            user: The user that is creating the built-in plugin parameter types.
            group: The group that will own the registered built-in plugin parameter
                types.
            commit: If True, commit the transaction. Defaults to True.

        Returns:
            A list of the newly registered built-in plugin parameter types.
        """
        log: BoundLogger = kwargs.get("log", LOGGER.new())
        builtin_types = list(BUILTIN_TYPES.keys())
        new_builtin_parameter_types: list[models.PluginTaskParameterType] = []
        for builtin_type in builtin_types:
            resource = models.Resource(resource_type=RESOURCE_TYPE, owner=group)
            new_plugin_parameter_type = models.PluginTaskParameterType(
                name=builtin_type,
                structure=None,
                description=None,
                resource=resource,
                creator=user,
            )
            db.session.add(new_plugin_parameter_type)
            db.session.add(
                models.ResourceLock(
                    resource_lock_type=resource_lock_types.READONLY,
                    resource=resource,
                )
            )
            new_builtin_parameter_types.append(new_plugin_parameter_type)

        if commit:
            db.session.commit()
            log.debug(
                "Built-in Plugin Parameter Types registration successful",
                builtin_parameter_type_ids=[
                    x.resource_id for x in new_builtin_parameter_types
                ],
            )

        return [
            utils.PluginParameterTypeDict(
                plugin_task_parameter_type=new_builtin_parameter_type, has_draft=False
            )
            for new_builtin_parameter_type in new_builtin_parameter_types
        ]



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import structlog
from sqlalchemy import select
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import db, models
from dioptra.restapi.errors import EntityDoesNotExistError
from dioptra.restapi.v1.entrypoints.service import (
    RESOURCE_TYPE as ENTRYPONT_RESOURCE_TYPE,
)
from dioptra.restapi.v1.experiments.service import (
    RESOURCE_TYPE as EXPERIMENT_RESOURCE_TYPE,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def get_entry_point(
    job_id: int, logger: BoundLogger | None = None
) -> models.EntryPoint:
    """Run a query to get the entrypoint for a job.

    Args:
        job_id: The ID of the job to get the entrypoint for.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The entrypoint for the job.
    """
    log = logger or LOGGER.new()  # noqa: F841

    entry_point_stmt = (
        select(models.EntryPoint)
        .join(models.EntryPointJob)
        .where(models.EntryPointJob.job_resource_id == job_id)
    )
    entry_point = db.session.scalar(entry_point_stmt)

    if entry_point is None:
        raise EntityDoesNotExistError(ENTRYPONT_RESOURCE_TYPE, job_id=job_id)

    return entry_point


def get_experiment(job_id: int, logger: BoundLogger | None = None) -> models.Experiment:
    """Run a query to get the experiment containing the job.

    Args:
        job_id: The ID of the job to get the experiment of.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The experiment containing the job.
    """
    log = logger or LOGGER.new()  # noqa: F841

    experiment_stmt = (
        select(models.Experiment)
        .join(models.ExperimentJob)
        .where(models.ExperimentJob.job_resource_id == job_id)
    )
    experiment = db.session.scalar(experiment_stmt)

    if experiment is None:
        raise EntityDoesNotExistError(EXPERIMENT_RESOURCE_TYPE, job_id=job_id)

    return experiment


def get_entry_point_plugin_files(
    job_id: int, logger: BoundLogger | None = None
) -> list[models.EntryPointPluginFile]:
    """Run a query to get the plugin files for an entrypoint.

    Args:
        job_id: The ID of the job to get the plugin files for.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The plugin files for the entrypoint.
    """
    log = logger or LOGGER.new()  # noqa: F841

    entry_point_resource_snapshot_id_stmt = (
        select(models.EntryPoint.resource_snapshot_id)
        .join(models.EntryPointJob)
        .where(
            models.EntryPointJob.job_resource_id == job_id,
        )
    )
    entry_point_plugin_files_stmt = select(models.EntryPointPluginFile).where(
        models.EntryPointPluginFile.entry_point_resource_snapshot_id
        == entry_point_resource_snapshot_id_stmt.scalar_subquery(),
    )
    return list(db.session.scalars(entry_point_plugin_files_stmt).unique().all())


def get_job_parameter_values(
    job_id: int, logger: BoundLogger | None = None
) -> list[models.EntryPointParameterValue]:
    """Run a query to get the parameter values for the job.

    Args:
        job_id: The ID of the job to get the parameter values for.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The parameter values for the job.
    """
    log = logger or LOGGER.new()  # noqa: F841

    entry_point_param_values_stmt = select(models.EntryPointParameterValue).where(
        models.EntryPointParameterValue.job_resource_id == job_id,
    )
    return list(db.session.scalars(entry_point_param_values_stmt).unique().all())


def get_plugin_parameter_types(
    job_id: int, logger: BoundLogger | None = None
) -> list[models.PluginTaskParameterType]:
    """Run a query to get the plugin task parameter types for the job.

    Args:
        job_id: The ID of the job to get the plugin task parameter types for.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The plugin files for the entrypoint.
    """
    log = logger or LOGGER.new()  # noqa: F841

    group_id_stmt = select(models.Resource.group_id).where(
        models.Resource.resource_id == job_id
    )
    plugin_parameter_types_stmt = (
        select(models.PluginTaskParameterType)
        .join(models.Resource)
        .where(
            models.Resource.is_deleted == False,  # noqa: E712
            models.Resource.group_id == group_id_stmt.scalar_subquery(),
            models.Resource.latest_snapshot_id
            == models.PluginTaskParameterType.resource_snapshot_id,
        )
    )
    return list(db.session.scalars(plugin_parameter_types_stmt).all())



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, ForeignKeyConstraint, Index, Text, select
from sqlalchemy.orm import Mapped, column_property, mapped_column, relationship

from dioptra.restapi.db.db import bigint, db, guid, intpk, text_

from .entry_points import EntryPoint
from .experiments import Experiment
from .queues import Queue
from .resources import ResourceSnapshot

if TYPE_CHECKING:
    from .entry_points import EntryPointParameterValue
    from .resources import Resource

# -- Tables (no ORM) -------------------------------------------------------------------

job_status_types_table = db.Table(
    "job_status_types",
    Column("status", Text(), primary_key=True),
)

# -- ORM Classes -----------------------------------------------------------------------


class EntryPointJob(db.Model):  # type: ignore[name-defined]
    __tablename__ = "entry_point_jobs"

    # Database fields
    entry_point_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("entry_points.resource_snapshot_id"), init=False
    )
    job_resource_id: Mapped[intpk] = mapped_column(
        ForeignKey("resources.resource_id"), init=False
    )

    # Derived fields (read-only)
    entry_point_id: Mapped[bigint] = column_property(
        select(EntryPoint.resource_id)
        .where(EntryPoint.resource_snapshot_id == entry_point_resource_snapshot_id)
        .correlate_except(EntryPoint)
        .scalar_subquery()
    )

    # Relationships
    job_resource: Mapped["Resource"] = relationship()
    entry_point: Mapped["EntryPoint"] = relationship(
        back_populates="entry_point_jobs", lazy="joined"
    )
    jobs: Mapped[list["Job"]] = relationship(
        init=False, back_populates="entry_point_job"
    )
    entry_point_parameter_values: Mapped[list["EntryPointParameterValue"]] = (
        relationship(
            back_populates="entry_point_job", overlaps="job_resource,parameter,values"
        )
    )

    # Additional settings
    __table_args__ = (Index(None, "job_resource_id", unique=True),)


class ExperimentJob(db.Model):  # type: ignore[name-defined]
    __tablename__ = "experiment_jobs"

    # Database fields
    experiment_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("experiments.resource_snapshot_id"), init=False
    )
    job_resource_id: Mapped[intpk] = mapped_column(
        ForeignKey("resources.resource_id"), init=False
    )

    # Derived fields (read-only)
    experiment_id: Mapped[bigint] = column_property(
        select(Experiment.resource_id)
        .where(Experiment.resource_snapshot_id == experiment_resource_snapshot_id)
        .correlate_except(Experiment)
        .scalar_subquery()
    )

    # Relationships
    job_resource: Mapped["Resource"] = relationship()
    experiment: Mapped["Experiment"] = relationship(
        back_populates="experiment_jobs", lazy="joined"
    )
    jobs: Mapped[list["Job"]] = relationship(
        init=False, back_populates="experiment_job", overlaps="jobs"
    )

    # Additional settings
    __table_args__ = (Index(None, "job_resource_id", unique=True),)


class QueueJob(db.Model):  # type: ignore[name-defined]
    __tablename__ = "queue_jobs"

    # Database fields
    queue_resource_snapshot_id: Mapped[intpk] = mapped_column(
        ForeignKey("queues.resource_snapshot_id"), init=False
    )
    job_resource_id: Mapped[intpk] = mapped_column(
        ForeignKey("resources.resource_id"), init=False
    )

    # Derived fields (read-only)
    queue_id: Mapped[bigint] = column_property(
        select(Queue.resource_id)
        .where(Queue.resource_snapshot_id == queue_resource_snapshot_id)
        .correlate_except(Queue)
        .scalar_subquery()
    )

    # Relationships
    job_resource: Mapped["Resource"] = relationship()
    queue: Mapped["Queue"] = relationship(back_populates="queue_jobs", lazy="joined")
    jobs: Mapped[list["Job"]] = relationship(
        init=False, back_populates="queue_job", overlaps="jobs"
    )

    # Additional settings
    __table_args__ = (Index(None, "job_resource_id", unique=True),)


class JobMlflowRun(db.Model):  # type: ignore[name-defined]
    __tablename__ = "job_mlflow_runs"

    # Database fields
    job_resource_id: Mapped[intpk] = mapped_column(ForeignKey("resources.resource_id"))
    mlflow_run_id: Mapped[guid] = mapped_column(nullable=False)

    # Relationships
    job_resource: Mapped["Resource"] = relationship(init=False)
    jobs: Mapped[list["Job"]] = relationship(
        init=False,
        back_populates="mlflow_run",
        primaryjoin="JobMlflowRun.job_resource_id == foreign(Job.resource_id)",
        overlaps="jobs",
    )

    # Additional settings
    __table_args__ = (Index(None, "mlflow_run_id", unique=True),)


class Job(ResourceSnapshot):
    __tablename__ = "jobs"

    # Database fields
    resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    resource_id: Mapped[bigint] = mapped_column(
        ForeignKey("entry_point_jobs.job_resource_id"),
        ForeignKey("experiment_jobs.job_resource_id"),
        ForeignKey("queue_jobs.job_resource_id"),
        init=False,
        nullable=False,
        index=True,
    )
    timeout: Mapped[text_] = mapped_column(nullable=False)
    status: Mapped[text_] = mapped_column(
        ForeignKey("job_status_types.status"), nullable=False, index=True
    )

    # Relationships
    entry_point_job: Mapped["EntryPointJob"] = relationship(
        init=False, back_populates="jobs", overlaps="jobs"
    )
    experiment_job: Mapped["ExperimentJob"] = relationship(
        init=False, back_populates="jobs", overlaps="entry_point_job,jobs"
    )
    mlflow_run: Mapped["JobMlflowRun"] = relationship(
        init=False,
        back_populates="jobs",
        primaryjoin="JobMlflowRun.job_resource_id == foreign(Job.resource_id)",
        overlaps="entry_point_job,experiment_job,jobs",
    )
    queue_job: Mapped["QueueJob"] = relationship(
        init=False,
        back_populates="jobs",
        overlaps="entry_point_job,experiment_job,mlflow_run,jobs",
    )

    # Additional settings
    __table_args__ = (  # type: ignore[assignment]
        Index(None, "resource_snapshot_id", "resource_id", unique=True),
        ForeignKeyConstraint(
            ["resource_snapshot_id", "resource_id"],
            [
                "resource_snapshots.resource_snapshot_id",
                "resource_snapshots.resource_id",
            ],
        ),
    )
    __mapper_args__ = {
        "polymorphic_identity": "job",
    }



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
A module for Dioptra's REST API query language.

This module is responsible for defining the query language and providing a parser.
It parses syntactically correct search queries into a list of search terms. It also
provides support for constructing sqlalchemy WHERE clauses from a parsed query.
"""
from typing import Any

import pyparsing as pp
from sqlalchemy import and_, or_

from dioptra.restapi.errors import SearchParseError


def _define_query_grammar() -> pp.ParserElement:
    """
    Defines the grammar for the Dioptra query language. The value of the REST
    API's `search` query parameter conforms to this grammar.

    The grammar supports a comma separated list of search terms.

    Returns:
        The pyparsing grammar.
    """

    # A wildcard is a '*' (multi) or a '?' (single) that is not escaped with a '\'
    wildcard = ~pp.Literal("\\") + (pp.Literal("*") | pp.Literal("?"))

    # An escaped character is a '\' followed by a character that needs to be escaped
    escape = pp.Literal("\\") + pp.Word("\\?*\"'n", exact=1)

    # Spaces are allowed in unquoted searches, but not when a field is specified
    space = pp.Literal(" ")

    # Unquoted searches are limited to words containing alphanumerics and underscores
    unquoted_word = pp.Combine((escape | pp.Word(pp.alphanums + "_") + ~space)[1, ...])
    unquoted_words = (
        pp.Combine((escape | pp.Word(pp.alphanums + "_"))[1, ...]) | space
    )[1, ...]
    # quoted searches can include any printable characters
    sgl_quoted_words = pp.Combine(
        (escape | pp.Word(pp.printables + " ", exclude_chars="\\?*'\n"))[1, ...]
    )
    dbl_quoted_words = pp.Combine(
        (escape | pp.Word(pp.printables + " ", exclude_chars='\\?*"\n'))[1, ...]
    )

    # The search strings are a sequence of one or more valid characters and wildcards
    unquoted_search_word = (unquoted_word | wildcard)[1, ...]
    unquoted_search_string = (unquoted_words | wildcard)[1, ...]
    sgl_quoted_search_string = (
        pp.Suppress("'") + (sgl_quoted_words | wildcard)[1, ...] + pp.Suppress("'")
    )
    dbl_quoted_search_string = (
        pp.Suppress('"') + (dbl_quoted_words | wildcard)[1, ...] + pp.Suppress('"')
    )
    quoted_search_string = sgl_quoted_search_string | dbl_quoted_search_string
    search_string = pp.Group(quoted_search_string | unquoted_search_string)
    field_search_value = pp.Group(quoted_search_string | unquoted_search_word)

    # A field name can contain alphabetical characters and underscores.
    field_name = pp.Word(pp.alphas + "_")
    # A colon is used as the separator between search fields and search strings.
    assign = pp.Suppress(pp.Literal(":"))
    # A field search string is a field name and search string separated by a colon.
    field_search_string = pp.Group(field_name + assign + field_search_value)

    # A search term is either a single search string or a search field:string pair.
    search_term = field_search_string | search_string

    # The full grammar is a comma-delimited list of search terms.
    grammar = pp.delimitedList(search_term, delim=",")

    return grammar


DIOPTRA_QUERY_GRAMMAR = _define_query_grammar()


def parse_search_text(search_text: str) -> list[dict]:
    """
    Parses the search text into a tokenized list of search terms.

    Args:
        search_text: the raw search text provided by the user.

    Returns:
        A list of dictionaries containing the tokenized search terms. Each dictionary
            contains a 'field' and 'value' key. The field is the name of the field to
            be searched or None to indicate the query should not be restricted by
            field. The value is a list of strings that represent the search value.
    """

    parsed_search = DIOPTRA_QUERY_GRAMMAR.parse_string(
        search_text, parse_all=True
    ).as_list()
    formatted_result = []
    for term in parsed_search:
        if len(term) > 1 and isinstance(term[1], list):
            formatted_result.append({"field": term[0], "value": term[1]})
        else:
            formatted_result.append({"field": None, "value": term})
    return formatted_result


def construct_sql_search_value(search_term: list[str], fuzzy: bool = False) -> str:
    """
    Constructs a search value for a SQL query from a tokenized list by replacing
    wildcards and joining the string.

    Args:
        search_value: A tokenized list of strings representing a search term.
        fuzzy: Perform a fuzzy search by placing a wildcard between words.

    Returns:
        A string to be used as the value in the WHERE clause of a SQL query.
    """
    search_term = [text.replace("/", r"//") for text in search_term]
    search_term = [text.replace("%", r"/%") for text in search_term]
    search_term = [text.replace("_", r"/_") for text in search_term]
    search_term = ["%" if text == "*" else text for text in search_term]
    search_term = ["_" if text == "?" else text for text in search_term]
    search_term = [text.replace(r"\\", "\\") for text in search_term]
    search_term = [text.replace(r"\*", "*") for text in search_term]
    search_term = [text.replace(r"\?", "?") for text in search_term]
    search_term = [text.replace(r"\"", '"') for text in search_term]
    search_term = [text.replace(r"'", "'") for text in search_term]
    search_term = [text.replace(r"\n", "\n") for text in search_term]
    if fuzzy:
        return "%" + "%".join(search_term) + "%"
    else:
        return "".join(search_term)


def construct_sql_query_filters(search_string: str, searchable_fields: dict[str, Any]):
    """
    Constructs a search filter to be used by sqlalchemy.

    Args:
        search_string: A string conforming to the search grammar.

    Returns:
        A filter that can be used in a sqlalchemy query.

    Raises:
        SearchParseError: If a search string cannot be parsed.
    """
    if not search_string:
        return True

    try:
        parsed_search_terms = parse_search_text(search_string)
    except pp.ParseException as error:
        raise SearchParseError(error.line, repr(error)) from error

    query_filters: list = []
    for search_term in parsed_search_terms:
        field: str = search_term["field"]
        if field is None:
            value = construct_sql_search_value(search_term["value"], fuzzy=True)
            filter = or_(filter_fn(value) for filter_fn in searchable_fields.values())
        elif field in searchable_fields:
            value = construct_sql_search_value(search_term["value"])
            filter = searchable_fields[field](value)
        else:
            raise SearchParseError(search_string, f"'{field}' is not a valid field")
        query_filters.append(filter)

    return and_(*query_filters)


if __name__ == "__main__":
    """
    A simple demonstration of the query grammar with sample queries.
    """

    DIOPTRA_QUERY_GRAMMAR.run_tests(
        r"""
        # search all text fields matching 'search_*'
        search_*
        # search all text the exactly match 'search_*'
        search_\*
        # a quoted search term can contain spaces and other characters
        "search \"this\", and 'that'"
        # multiple search terms can be provided via a comma-separated list
        "search=this", 'and, how about "this\?"',this_too
        # search tags whose name matches 'classification' exactly
        tag:classification
        # search tags whose name starts with 'class'
        tag:class*
        # search descriptions containing mnist and whose tag name matches 'cv' exactly
        description:*mnist*,tag:cv
        # search for name starting with 'trial_' and ending with two valid characters
        name:trial_??
        # search for literal '*'
        \*
        # multi-word search
        search all for this
        # field multi-word search
        field:"search all for this"
        """,
        full_dump=False,
    )

    DIOPTRA_QUERY_GRAMMAR.run_tests(
        r"""
        # invalid multi word field search
        field:search all for this
        # incorrect assignment character used
        bad=assignment
        # invalid characters in field name
        bad-name:classification
        # invalid delimiter between search terms
        description:bad_delim|tag:cv
        # missing closing quote
        "forgot to close this quote
        # tried to escape an unescapable character
        \q
        """,
        full_dump=False,
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""
Define classes related to types.
"""

import enum
from collections.abc import Iterable, Mapping, Sequence, Set
from typing import Any, Optional, Union

# This creates a circular import.  The exception module imports this module
# because it needs these type classes as annotations.  I would like exception
# instances related to types to be able to store the relevant type objects.
# This module needs to raise exceptions, so it has to import the exceptions
# module.  So each module needs to import the other.  So we need to be careful
# how each module uses the other.  A "from X import Y" style import will not
# work here.
import dioptra.sdk.exceptions.task_engine


class StructureType(enum.Enum):
    """
    Define type structures, for structured types.
    """

    MAPPING = enum.auto()
    LIST = enum.auto()
    TUPLE = enum.auto()


def _make_iterable_hashable(iterable: Iterable[Any]) -> tuple[Any, ...]:
    """
    Make an arbitrary iterable into a hashable value.  The hash will be
    sensitive to value order within the iterable.

    Args:
        iterable: The iterable to convert

    Returns:
        A tuple of hashable values which represents the given iterable
    """

    hashable = tuple(_make_hashable(elt) for elt in iterable)

    return hashable


def _make_mapping_hashable(mapping: Mapping[Any, Any]) -> frozenset[Any]:
    """
    Make an arbitrary mapping into a hashable value.  The hash will be
    insensitive to any inherent entry ordering within the mapping.

    Args:
        mapping: The mapping to convert

    Returns:
        A frozenset of 2-tuples, which represents the given mapping
    """
    # assume keys are hashable (probably strings).  Using a frozenset ensures
    # the hash function is agnostic to an iteration order.  We want it to
    # depend only on the keys and values.  That's also how dict equality works.
    hashable = frozenset((key, _make_hashable(value)) for key, value in mapping.items())

    return hashable


def _make_hashable(value: Any) -> Any:
    """
    Make a hashable value from the given value.  This will transform the value
    to something else, which is representative of the original but hashable,
    if the value isn't already hashable.

    Args:
        value: The value to convert

    Returns:
        A hashable value
    """
    result: Any
    if isinstance(value, set):
        result = frozenset(value)  # just in case?
    elif isinstance(value, Mapping):
        result = _make_mapping_hashable(value)
    elif isinstance(value, Iterable) and not isinstance(value, str):
        result = _make_iterable_hashable(value)
    else:
        # else, assume hashability.
        result = value

    return result


class TypeStructure:
    """
    Instances represent the required structure of a structured type.
    """

    def __init__(
        self,
        struct_type: StructureType,
        struct_def: Union["Type", Mapping[str, "Type"], Sequence["Type"]],
    ) -> None:
        """
        Initialize this instance.

        Args:
            struct_type: A StructureType enum value which represents the
                structure type
            struct_def: The structure definition, which must be compatible with
                the chosen structure type.  This is only lightly validated at
                runtime, but must be a Type instance for lists, a sequence of
                Type instances for tuples and key/value type mappings (sequence
                length must be 2 for the latter), and a string->Type mapping
                for enumerated mapping structures.
        """

        self.__struct_type = struct_type
        self.__struct_def = struct_def

        self.__check_structure_agrees_with_type()

    @property
    def struct_type(self) -> StructureType:
        """
        The type of structure
        """
        return self.__struct_type

    @property
    def struct_def(self) -> Union["Type", Mapping[str, "Type"], Sequence["Type"]]:
        """
        The structure definition
        """
        return self.__struct_def

    def __check_structure_agrees_with_type(self) -> None:
        """
        Do a light sanity check of the structure definition given against
        the structure type.
        """
        if self.struct_type is StructureType.MAPPING and not isinstance(
            self.struct_def, (Mapping, Sequence)
        ):
            raise dioptra.sdk.exceptions.task_engine.InvalidTypeStructureDefinitionError(  # noqa: B950
                "For mapping type structures, the definition must be"
                " either a mapping (e.g. a dict) from property names to types,"
                " or a 2-element list containing a key type followed by a value"
                " type"
            )

        elif self.struct_type is StructureType.LIST and not isinstance(
            self.struct_def, Type
        ):
            raise dioptra.sdk.exceptions.task_engine.InvalidTypeStructureDefinitionError(  # noqa: B950
                "For a list type structure, the definition must be a type"
            )

        elif self.struct_type is StructureType.TUPLE and not isinstance(
            self.struct_def, Sequence
        ):
            raise dioptra.sdk.exceptions.task_engine.InvalidTypeStructureDefinitionError(  # noqa: B950
                "For a tuple type structure, the definition must be a sequence"
                " of types"
            )

    def __eq__(self, other: Any) -> bool:
        """
        Check this type for equality against another.

        Args:
            other: The other value

        Returns:
            True if equal, False if not
        """

        result = False
        if self is other:
            result = True
        elif isinstance(other, TypeStructure):
            result = (
                self.struct_type is other.struct_type
                and self.struct_def == other.struct_def
            )
        # The above "==" check should automatically recurse inside structures.

        return result

    def __hash__(self) -> int:
        """
        Compute a hash value from this instance.

        Returns:
            The hash value
        """
        hashable_struct = _make_hashable(self.struct_def)

        return hash((self.struct_type, hashable_struct))

    def __repr__(self) -> str:
        """
        Create a readable repr string from this instance.

        Returns:
            The repr string
        """
        return "TypeStructure({!s}, {!r})".format(self.struct_type, self.struct_def)

    def __str__(self) -> str:
        """
        Create a readable string representation of this instance.

        Returns:
            The string representation
        """
        if self.struct_type is StructureType.MAPPING:
            if isinstance(self.struct_def, Mapping):
                # Enumerated key mapping structure
                struct_def_str = (
                    "{"
                    + ", ".join(
                        repr(k) + ": " + str(v) for k, v in self.struct_def.items()
                    )
                    + "}"
                )

            else:
                # Key type + value type mapping structure
                assert isinstance(self.struct_def, Sequence)  # for mypy
                struct_def_str = ", ".join(
                    str(elt_type) for elt_type in self.struct_def
                )

        elif self.struct_type is StructureType.LIST:
            struct_def_str = str(self.struct_def)

        else:  # tuple structure
            assert isinstance(self.struct_def, Sequence)  # for mypy
            struct_def_str = ", ".join(str(elt_type) for elt_type in self.struct_def)

        return self.struct_type.name + "[" + struct_def_str + "]"


class Type:
    """
    Immutable base class of all types.
    """

    def __init__(self, name: Optional[str] = None) -> None:
        """
        Initialize this type.

        Args:
            name: The type name
        """
        self.__name = name

    @property
    def name(self) -> Optional[str]:
        """
        The name of this type
        """
        return self.__name

    def __hash__(self) -> int:
        """
        Abstract __hash__ implementation; subclasses must override.
        """
        # Required so we can put types in sets
        raise NotImplementedError()

    def __eq__(self, other: Any) -> bool:
        """
        Abstract __eq__ implementation; subclasses must override.
        """
        raise NotImplementedError()


class SimpleType(Type):
    """
    Instances represent "simple" types: those which just have names.
    """

    def __init__(self, name: str, super_type: Optional["SimpleType"] = None) -> None:
        """
        Initialize this simple type.

        Args:
            name: The name of the type
            super_type: The super-type
        """

        if name is None:
            raise ValueError("Simple types must have a name")

        super().__init__(name)

        self.__super_type = super_type

    @property
    def name(self) -> str:
        """
        Overrides superclass "name" property to have "str" return type.  Simple
        types always have names (there is a check in the constructor), but the
        superclass property has type Optional[str] since other kinds of types
        may be anonymous (have no name).  I think it's easier to have this
        override in place, than to have to add asserts for mypy every single
        time I refer to the name of a simple type in a non-optional "str"
        context.

        Returns:
            The name of this type
        """
        name = super().name
        # Required for mypy: simple types always have names.
        assert name is not None
        return name

    @property
    def super_type(self) -> Optional["SimpleType"]:
        """
        The super-type
        """
        return self.__super_type

    def is_subtype_of(self, other: Any) -> bool:
        """
        Determine whether this type is a subtype of the other

        Args:
            other: A type

        Returns:
            True if self is a subtype of other; False if not
        """
        result = False
        type_ = self.super_type
        while type_ is not None:
            if type_ == other:
                result = True
                break
            type_ = type_.super_type

        return result

    def __eq__(self, other: Any) -> bool:
        """
        Determine whether self is equal to other.

        Args:
            other: Another value to compare to self

        Returns:
            True if self is equal to other; False if not
        """
        # If names equal, we treat the types as equal.  It is not legal to have
        # two types with the same name but which are different in some other
        # respect.  Treat the names as representative of everything else.
        # Simple types must have names; they can't be anonymous.
        result = False
        if self is other:
            result = True
        elif isinstance(other, SimpleType):
            result = self.name == other.name

        return result

    def __hash__(self) -> int:
        """
        Compute a hash value for this type

        Returns:
            The hash value
        """
        # See __eq__: name is uniquely representative
        return hash(self.name)

    def __repr__(self) -> str:
        """
        Create a readable repr string for this type.

        Returns:
            The repr string
        """
        return "SimpleType(name={!r}, super_type={!r})".format(
            self.name, self.super_type
        )

    def __str__(self) -> str:
        """
        Create a readable string representation of this type.

        Returns:
            The string representation
        """
        # Just name, to decrease verbosity
        return self.name


class StructuredType(Type):
    """
    Instances represent "structured" types, i.e. those with defined internal
    structure.
    """

    def __init__(self, structure: TypeStructure, name: Optional[str] = None) -> None:
        """
        Initialize this type.

        Args:
            structure: The structure of this type
            name: The name of this type
        """
        super().__init__(name)

        if structure is None:
            raise ValueError("Structured types must have a structure")

        self.__structure = structure

    @property
    def structure(self) -> TypeStructure:
        """
        The structure of this type
        """
        return self.__structure

    def __eq__(self, other: Any) -> bool:
        """
        Determine whether self is equal to other.

        Args:
            other: Another value to compare to self

        Returns:
            True if self is equal to other; False if not
        """
        result = False
        if self is other:
            result = True
        elif isinstance(other, StructuredType):
            result = self.name == other.name

            if result and self.name is None:
                result = self.structure == other.structure

            # else: if names equal and not None, we treat the types as equal.
            # It is not legal to have two types with the same name but which are
            # different in some other respect.  Treat the names as
            # representative of everything else.  If we have names, there is no
            # need to check anything else.

        return result

    def __hash__(self) -> int:
        """
        Compute a hash value for this type

        Returns:
            The hash value
        """
        if self.name is not None:
            result = hash(self.name)
        else:
            # anonymous: defined by its structure
            result = hash(self.structure)

        return result

    def __repr__(self) -> str:
        """
        Create a readable repr string for this type.

        Returns:
            The repr string
        """
        return "StructuredType(structure={!r}, name={!r})".format(
            self.structure, self.name
        )

    def __str__(self) -> str:
        """
        Create a readable string representation of this type.

        Returns:
            The string representation
        """
        s = "{}: {}".format(self.name or "<anon>", str(self.structure))

        return s


class UnionType(Type):
    """
    Instances represent union types.
    """

    def __init__(self, member_types: Iterable[Type], name=None) -> None:
        """
        Initialize this type.

        Args:
            member_types: The member types of the union
            name: The name of this type
        """
        super().__init__(name)

        # Use set to dedupe and obtain order-agnosticism
        self.__member_types = frozenset(member_types)

    @property
    def member_types(self) -> Set[Type]:
        """
        The member types of the union.
        """
        return self.__member_types

    def __eq__(self, other: Any) -> bool:
        """
        Determine whether self is equal to other.

        Args:
            other: Another value to compare to self

        Returns:
            True if self is equal to other; False if not
        """
        result = False
        if self is other:
            result = True
        elif isinstance(other, UnionType):
            result = self.name == other.name

            if result and self.name is None:
                result = self.member_types == other.member_types

            # else: if names equal and not None, we treat the types as equal.
            # It is not legal to have two types with the same name but which are
            # different in some other respect.  Treat the names as
            # representative of everything else.  If we have names, there is no
            # need to check anything else.

        return result

    def __hash__(self) -> int:
        """
        Compute a hash value for this type

        Returns:
            The hash value
        """
        if self.name is not None:
            result = hash(self.name)
        else:
            # anonymous: defined by its members
            result = hash(self.member_types)

        return result

    def __repr__(self) -> str:
        """
        Create a readable repr string for this type.

        Returns:
            The repr string
        """
        # I like the bracketed list look more than frozenset(...).
        members_list_repr = "[" + ", ".join(repr(m) for m in self.member_types) + "]"
        return "UnionType(member_types={}, name={!r})".format(
            members_list_repr, self.name
        )

    def __str__(self) -> str:
        """
        Create a readable string representation of this type.

        Returns:
            The string representation
        """
        s = "{}: UNION[{}]".format(
            self.name or "<anon>", ", ".join(str(t) for t in self.member_types)
        )

        return s
